# future
from __future__ import annotations

# stdlib
from collections import UserString
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import SupportsIndex

# syft absolute
import syft as sy

# relative
from ...core.common import UID
from ...core.common.serde.serializable import serializable
from ...proto.lib.python.string_pb2 import String as String_PB
from .int import Int
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .slice import Slice
from .types import SyPrimitiveRet


@serializable()
class String(UserString, PyPrimitive):
    def __init__(
        self,
        value: Any = None,
        id: Optional[UID] = None,
        temporary_box: bool = False,
    ):

        if value is None:
            value = ""

        UserString.__init__(self, value)
        PyPrimitive.__init__(self, temporary_box=temporary_box)

        self._id: UID = id if id else UID()

    def upcast(self) -> str:
        return str(self)

    def __add__(self, other: Any) -> SyPrimitiveRet:
        res = super().__add__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        res = super().__eq__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __float__(self) -> SyPrimitiveRet:
        res = super().__float__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __ge__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ge__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __getitem__(self, key: Union[int, slice, Slice, SupportsIndex]) -> Any:
        if isinstance(key, Slice):
            key = key.upcast()
        res = super().__getitem__(key)
        return PrimitiveFactory.generate_primitive(value=res)

    def __gt__(self, other: Any) -> SyPrimitiveRet:
        res = super().__gt__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __hash__(self) -> SyPrimitiveRet:
        res = super().__hash__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __int__(self) -> SyPrimitiveRet:
        res = super().__int__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __iter__(self) -> SyPrimitiveRet:
        # TODO fix this
        res = super().__iter__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __le__(self, other: Any) -> SyPrimitiveRet:
        res = super().__le__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __len__(self) -> SyPrimitiveRet:
        res = super().__len__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __lt__(self, other: Any) -> SyPrimitiveRet:
        res = super().__lt__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __mod__(self, *args: Any) -> SyPrimitiveRet:
        res = super().__mod__(
            *[str(arg) if isinstance(arg, String) else arg for arg in args]
        )
        return PrimitiveFactory.generate_primitive(value=res)

    def __mul__(self, other: Any) -> SyPrimitiveRet:
        res = super().__mul__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ne__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __reversed__(self) -> Any:
        # returns <class 'reversed'>
        return super().__reversed__()

    def __sizeof__(self) -> SyPrimitiveRet:
        res = super().__sizeof__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __str__(self) -> str:
        return super().__str__()

    def capitalize(self) -> SyPrimitiveRet:
        res = super().capitalize()
        return PrimitiveFactory.generate_primitive(value=res)

    def casefold(self) -> SyPrimitiveRet:
        res = super().casefold()
        return PrimitiveFactory.generate_primitive(value=res)

    def center(self, width: Union[int, Int], *args: Any) -> SyPrimitiveRet:
        if args:
            _args_0 = str(args[0]) if isinstance(args[0], String) else args[0]
            res = super().center(width, _args_0, *args[1:])
        else:
            res = super().center(width)
        return PrimitiveFactory.generate_primitive(value=res)

    def count(
        self, sub: Any, start: Optional[int] = None, end: Optional[int] = None
    ) -> SyPrimitiveRet:
        res = super().count(sub, start, end)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def encode(
        self, encoding: Optional[str] = None, errors: Optional[str] = None
    ) -> SyPrimitiveRet:
        res = super().encode(encoding, errors)
        return PrimitiveFactory.generate_primitive(value=res)

    def endswith(
        self,
        suffix: Union[str, "String", tuple],
        start: Optional[int] = None,
        end: Optional[int] = None,
    ) -> SyPrimitiveRet:
        suffix = str(suffix) if isinstance(suffix, String) else suffix
        _suffix = (
            tuple(str(elem) if isinstance(elem, String) else elem for elem in suffix)
            if isinstance(suffix, tuple)
            else suffix
        )
        res = super().endswith(_suffix, start, end)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def expandtabs(self, tabsize: int = 8) -> SyPrimitiveRet:
        res = super().expandtabs(tabsize)
        return PrimitiveFactory.generate_primitive(value=res)

    def find(
        self, sub: Any, start: Optional[int] = 0, end: Optional[int] = None
    ) -> SyPrimitiveRet:
        if end is None:
            end = super().__len__()

        res = super().find(sub, start, end)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def format(self, *args: Any, **kwargs: Any) -> SyPrimitiveRet:
        res = super().format(*args, **kwargs)
        return PrimitiveFactory.generate_primitive(value=res)

    def format_map(self, mapping: Mapping[Any, Any]) -> SyPrimitiveRet:
        res = super().format_map(mapping)
        return PrimitiveFactory.generate_primitive(value=res)

    def index(
        self,
        sub: Union[str, "String"],
        start: Optional[int] = 0,
        end: Optional[int] = None,
    ) -> SyPrimitiveRet:
        if end is None:
            end = super().__len__()
        res = super().index(str(sub), start, end)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def isalnum(self) -> SyPrimitiveRet:
        res = super().isalnum()
        return PrimitiveFactory.generate_primitive(value=res)

    def isascii(self) -> SyPrimitiveRet:
        res = super().isascii()  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def isalpha(self) -> SyPrimitiveRet:
        res = super().isalpha()
        return PrimitiveFactory.generate_primitive(value=res)

    def isdecimal(self) -> SyPrimitiveRet:
        res = super().isdecimal()
        return PrimitiveFactory.generate_primitive(value=res)

    def isdigit(self) -> SyPrimitiveRet:
        res = super().isdigit()
        return PrimitiveFactory.generate_primitive(value=res)

    def isidentifier(self) -> SyPrimitiveRet:
        res = super().isidentifier()
        return PrimitiveFactory.generate_primitive(value=res)

    def islower(self) -> SyPrimitiveRet:
        res = super().islower()
        return PrimitiveFactory.generate_primitive(value=res)

    def isnumeric(self) -> SyPrimitiveRet:
        res = super().isnumeric()
        return PrimitiveFactory.generate_primitive(value=res)

    def isprintable(self) -> SyPrimitiveRet:
        res = super().isprintable()
        return PrimitiveFactory.generate_primitive(value=res)

    def isspace(self) -> SyPrimitiveRet:
        res = super().isspace()
        return PrimitiveFactory.generate_primitive(value=res)

    def istitle(self) -> SyPrimitiveRet:
        res = super().istitle()
        return PrimitiveFactory.generate_primitive(value=res)

    def isupper(self) -> SyPrimitiveRet:
        res = super().isupper()
        return PrimitiveFactory.generate_primitive(value=res)

    def join(self, seq: Any) -> SyPrimitiveRet:
        res = super().join(
            [str(elem) if isinstance(elem, String) else elem for elem in seq]
        )
        return PrimitiveFactory.generate_primitive(value=res)

    def ljust(self, width: Union[int], *args: Any) -> SyPrimitiveRet:
        if args:
            _args_0 = str(args[0]) if isinstance(args[0], String) else args[0]
            res = super().ljust(width, _args_0, *args[1:])
        else:
            res = super().ljust(width)
        return PrimitiveFactory.generate_primitive(value=res)

    def lower(self) -> SyPrimitiveRet:
        res = super().lower()
        return PrimitiveFactory.generate_primitive(value=res)

    def lstrip(self, chars: Optional[Union[str, "String"]] = None) -> SyPrimitiveRet:
        chars = str(chars) if isinstance(chars, String) else chars
        res = super().lstrip(chars)
        return PrimitiveFactory.generate_primitive(value=res)

    def partition(self, sep: Optional[Union[str, "String"]] = " ") -> SyPrimitiveRet:
        sep = str(sep) if isinstance(sep, String) else sep
        res = super().partition(sep)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def replace(
        self,
        oldvalue: Union[str, UserString],
        newvalue: Union[str, UserString],
        count: Optional[int] = -1,
    ) -> SyPrimitiveRet:
        res = super().replace(str(oldvalue), str(newvalue), count)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def rfind(
        self,
        sub: Union[str, UserString],
        start: Optional[int] = 0,
        end: Optional[int] = None,
    ) -> SyPrimitiveRet:
        sub = str(sub) if isinstance(sub, UserString) else sub
        res = super().rfind(sub, start, end)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def rindex(
        self,
        sub: Union[str, UserString],
        start: Optional[int] = 0,
        end: Optional[int] = None,
    ) -> SyPrimitiveRet:
        sub = str(sub) if isinstance(sub, String) else sub
        res = super().rindex(sub, start, end)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def rjust(self, width: int, *args: Any) -> SyPrimitiveRet:
        if args:
            _args_0 = str(args[0]) if isinstance(args[0], String) else args[0]
            res = super().rjust(width, _args_0, *args[1:])
        else:
            res = super().rjust(width)
        return PrimitiveFactory.generate_primitive(value=res)

    def rpartition(self, sep: Optional[Union[str, "String"]] = " ") -> SyPrimitiveRet:
        sep = str(sep) if isinstance(sep, String) else sep
        res = super().rpartition(sep)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def rsplit(
        self, sep: Optional[Union[str, "String"]] = None, maxsplit: int = -1
    ) -> SyPrimitiveRet:
        sep = str(sep) if isinstance(sep, String) else sep
        res = super().rsplit(sep=sep, maxsplit=maxsplit)
        return PrimitiveFactory.generate_primitive(value=res)

    def rstrip(self, chars: Optional[Union[str, "String"]] = None) -> SyPrimitiveRet:
        chars = str(chars) if isinstance(chars, String) else chars
        res = super().rstrip(chars)
        return PrimitiveFactory.generate_primitive(value=res)

    def split(
        self, sep: Optional[Union[str, "String"]] = None, maxsplit: int = -1
    ) -> SyPrimitiveRet:
        sep = str(sep) if isinstance(sep, String) else sep
        res = super().split(sep=sep, maxsplit=maxsplit)
        return PrimitiveFactory.generate_primitive(value=res)

    def splitlines(self, keepends: bool = False) -> SyPrimitiveRet:
        res = super().splitlines(keepends)
        return PrimitiveFactory.generate_primitive(value=res)

    def startswith(
        self,
        suffix: Union[str, UserString, tuple],
        start: Optional[int] = None,
        end: Optional[int] = None,
    ) -> SyPrimitiveRet:
        suffix = str(suffix) if isinstance(suffix, UserString) else suffix
        suffix = (
            tuple(
                str(elem) if isinstance(elem, UserString) else elem for elem in suffix
            )
            if isinstance(suffix, tuple)
            else suffix
        )
        start = start if start else 0
        end = end if end else len(self)
        res = super().startswith(suffix, start, end)
        return PrimitiveFactory.generate_primitive(value=res)

    def strip(self, chars: Optional[str] = None) -> SyPrimitiveRet:
        chars = str(chars) if isinstance(chars, String) else chars  # type: ignore
        res = super().strip(chars)
        return PrimitiveFactory.generate_primitive(value=res)

    def swapcase(self) -> SyPrimitiveRet:
        res = super().swapcase()
        return PrimitiveFactory.generate_primitive(value=res)

    def title(self) -> SyPrimitiveRet:
        res = super().title()
        return PrimitiveFactory.generate_primitive(value=res)

    def translate(self, *args: Any) -> SyPrimitiveRet:
        res = super().translate(*args)
        return PrimitiveFactory.generate_primitive(value=res)

    def upper(self) -> SyPrimitiveRet:
        res = super().upper()
        return PrimitiveFactory.generate_primitive(value=res)

    def zfill(self, width: Union[int, Int]) -> SyPrimitiveRet:
        res = super().zfill(width)
        return PrimitiveFactory.generate_primitive(value=res)

    def __contains__(self, val: object) -> SyPrimitiveRet:
        res = super().__contains__(val)
        return PrimitiveFactory.generate_primitive(value=res)

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def _object2proto(self) -> String_PB:
        return String_PB(
            data=self.data,
            id=sy.serialize(obj=self.id),
            temporary_box=self.temporary_box,
        )

    @staticmethod
    def _proto2object(proto: String_PB) -> "String":
        str_id: UID = sy.deserialize(blob=proto.id)
        return String(
            value=proto.data,
            id=str_id,
            temporary_box=proto.temporary_box,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return String_PB

    # fixes __rmod__ in python <= 3.7
    # https://github.com/python/cpython/commit/7abf8c60819d5749e6225b371df51a9c5f1ea8e9
    def __rmod__(
        self, template: Union[PyPrimitive, str, object]
    ) -> Union[SyPrimitiveRet, String]:
        return self.__class__(str(template) % self)



# stdlib
import logging
import sys
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

# third party
import pandas as pd

# relative
from ..... import deserialize
from .....core.tensor.autodp.adp_tensor import ADPTensor
from .....core.tensor.tensor import Tensor
from ....common import UID
from ....common.serde.serialize import _serialize as serialize  # noqa: F401
from ...abstract.node import AbstractNodeClient
from ...enums import RequestAPIFields
from ...enums import ResponseObjectEnum
from ..node_service.dataset_manager.dataset_manager_messages import CreateDatasetMessage
from ..node_service.dataset_manager.dataset_manager_messages import DeleteDatasetMessage
from ..node_service.dataset_manager.dataset_manager_messages import GetDatasetMessage
from ..node_service.dataset_manager.dataset_manager_messages import GetDatasetsMessage
from ..node_service.dataset_manager.dataset_manager_messages import UpdateDatasetMessage
from .request_api import RequestAPI

initial_boilerplate = """<style>
        #myInput {
          background-position: 10px 12px; /* Position the search icon */
          background-repeat: no-repeat; /* Do not repeat the icon image */
          background-color: #bbb;
          width: 98%; /* Full-width */
          font-size: 14px; /* Increase font-size */
          padding: 12px 20px 12px 40px; /* Add some padding */
          border: 1px solid #ddd; /* Add a grey border */
          margin-bottom: 12px; /* Add some space below the input */
        }

        #myTable {
          border-collapse: collapse; /* Collapse borders */
          width: 50%; /* Full-width */
          border: 1px solid #ddd; /* Add a grey border */
          font-size: 14px; /* Increase font-size */
        }

        #myTable th, #myTable td {
          text-align: left; /* Left-align text */
          padding: 10px; /* Add padding */
        }

        #myTable tr {
          /* Add a bottom border to all table rows */
          border-bottom: 1px solid #ddd;
        }

        #myTable tr.header, #myTable tr:hover {
          /* Add a grey background color to the table header and on hover */
          background-color: #777;
        }
        </style>

        <table id="myTable">
          <tr class="header">
            <th style="width:15%;">Asset Key</th>
            <th style="width:20%;">Type</th>
            <th style="width:10%;">Shape</th>
          </tr>
        """

end_boilerplate = """
        </table>

        <script>
        function myFunction() {
          // Declare variables
          var input, filter, table, tr, td, i, txtValue;
          input = document.getElementById("myInput");
          filter = input.value.toUpperCase();
          table = document.getElementById("myTable");
          tr = table.getElementsByTagName("tr");

          // Loop through all table rows, and hide those who don't match the search query
          for (i = 0; i < tr.length; i++) {
            name_td = tr[i].getElementsByTagName("td")[1];
            desc_td = tr[i].getElementsByTagName("td")[2];
            asset_td = tr[i].getElementsByTagName("td")[3];
            id_td = tr[i].getElementsByTagName("td")[4];
            if (name_td || desc_td || asset_td || id_td) {
              name_txtValue = name_td.textContent || name_td.innerText;
              desc_txtValue = desc_td.textContent || name_td.innerText;
              asset_txtValue = asset_td.textContent || name_td.innerText;
              id_txtValue = id_td.textContent || name_td.innerText;
              name_bool = name_txtValue.toUpperCase().indexOf(filter) > -1;
              desc_bool = desc_txtValue.toUpperCase().indexOf(filter) > -1;
              asset_bool = asset_txtValue.toUpperCase().indexOf(filter) > -1;
              id_bool = id_txtValue.toUpperCase().indexOf(filter) > -1;
              if (name_bool || desc_bool || asset_bool || id_bool) {
                tr[i].style.display = "";
              } else {
                tr[i].style.display = "none";
              }
            }
          }
        }
        </script>"""


class DatasetRequestAPI(RequestAPI):
    def __init__(self, client: AbstractNodeClient):
        super().__init__(
            client=client,
            create_msg=CreateDatasetMessage,
            get_msg=GetDatasetMessage,
            get_all_msg=GetDatasetsMessage,
            update_msg=UpdateDatasetMessage,
            delete_msg=DeleteDatasetMessage,
            response_key=ResponseObjectEnum.DATASET,
        )

    def create_syft(self, **kwargs: Any) -> None:
        super().create(**kwargs)

    def create_grid_ui(self, path: str, **kwargs) -> Dict[str, str]:  # type: ignore
        response = self.node.conn.send_files(  # type: ignore
            "/datasets", path, form_name="metadata", form_values=kwargs
        )  # type: ignore
        logging.info(response[RequestAPIFields.MESSAGE])

    def all(self) -> List[Any]:
        result = [
            content
            for content in self.perform_api_request(
                syft_msg=self._get_all_message, timeout=1
            ).metadatas
        ]

        new_all = list()
        for dataset in result:
            new_dataset = {}
            for k, v_blob in dataset.items():
                if k not in ["str_metadata", "blob_metadata", "manifest"]:
                    new_dataset[k] = deserialize(v_blob, from_bytes=True)
            new_all.append(new_dataset)

        return new_all

    def __getitem__(self, key: Union[str, int, slice]) -> Any:
        # optionally we should be able to pass in the index of the dataset we want
        # according to the order displayed when displayed as a pandas table

        if isinstance(key, int):
            a = self.all()
            d = a[key : key + 1]  # noqa: E203
            return Dataset(d, self.client, key=key, **a[key])

        elif isinstance(key, slice):

            class NewObject:
                def _repr_html_(self2: Any) -> str:
                    return self.dataset_list_to_html(
                        self.all_as_datasets().__getitem__(key)  # type: ignore
                    )

            return NewObject()

    def all_as_datasets(self) -> List[Any]:
        a = self.all()
        out = list()
        for key, _ in enumerate(a):
            raw = a[key : key + 1]  # noqa: E203
            out.append(Dataset(raw, self.client, key=key, **a[key]))
        return out

    def purge(self, skip_check: bool = False) -> None:
        if not skip_check:
            pref = input(
                "You are about to delete all datasets ? 🚨 \n"
                "All information will be permanantely deleted.\n"
                "Please enter y/n to proceed: "
            )
            while pref != "y" and pref != "n":
                pref = input("Invalid input '" + pref + "', please specify 'y' or 'n'.")
            if pref == "n":
                print("Datasets deletion is cancelled.")
                return None

        for dataset in self.all():
            self.delete(dataset_id=dataset.get("id"))

    def __len__(self) -> int:
        return len(self.all())

    def __delitem__(self, key: int) -> Any:

        try:
            dataset = self.all()[key]
        except IndexError as err:
            raise err

        dataset_id = dataset.get("id")
        dataset_name = dataset.get("name", "")

        pref = input(
            f"You are about to delete the `{dataset_name}` ? 🚨 \n"
            "All information related to this dataset will be permanantely deleted.\n"
            "Please enter y/n to proceed: "
        )
        while pref != "y" and pref != "n":
            pref = input("Invalid input '" + pref + "', please specify 'y' or 'n'.")
        if pref == "n":
            raise Exception("Dataset deletion is cancelled.")

        self.delete(dataset_id=dataset_id)
        sys.stdout.write(f"Dataset: `{dataset_name}` is successfully deleted.")

        return True

    def _repr_html_(self) -> str:
        if len(self) > 0:
            return self.dataset_list_to_html(self.all_as_datasets())
        return "(no datasets found)"

    @staticmethod
    def dataset_list_to_html(dataset_iterable: List[Any]) -> str:

        initial_boilerplate = """<style>
                #myInput {
                  background-position: 10px 12px; /* Position the search icon */
                  background-repeat: no-repeat; /* Do not repeat the icon image */
                  background-color: #bbb;
                  width: 98%; /* Full-width */
                  font-size: 14px; /* Increase font-size */
                  padding: 12px 20px 12px 40px; /* Add some padding */
                  border: 1px solid #ddd; /* Add a grey border */
                  margin-bottom: 12px; /* Add some space below the input */
                }

                #myTable {
                  border-collapse: collapse; /* Collapse borders */
                  width: 100%; /* Full-width */
                  border: 1px solid #ddd; /* Add a grey border */
                  font-size: 14px; /* Increase font-size */
                }

                #myTable th, #myTable td {
                  text-align: left; /* Left-align text */
                  padding: 10px; /* Add padding */
                }

                #myTable tr {
                  /* Add a bottom border to all table rows */
                  border-bottom: 1px solid #ddd;
                }

                #myTable tr.header, #myTable tr:hover {
                  /* Add a grey background color to the table header and on hover */
                  background-color: #777;
                }
                </style>

                <table id="myTable" style="width:1000px">
                  <tr class="header">
                    <th style="width:30px">Idx</th>
                    <th style="width:20%;">Name</th>
                    <th style="width:35%;">Description</th>
                    <th style="width:20%;">Assets</th>
                    <th style="width:300px;">Id</th>
                  </tr>
                """

        rows = ""
        for _, d in enumerate(dataset_iterable):

            data = d.data
            truncated_assets = False
            if len(data) > 3:
                truncated_assets = True
                data = data[:3]

            assets = ""
            for _, a in enumerate(data):
                assets += '["' + a["name"] + '"] -> ' + a["dtype"] + "<br /><br />"

            if truncated_assets:
                assets += "...<br /><br />"

            rows += (
                """

          <tr>
            <td>["""
                + str(d.key)
                + """]</td>
            <td>"""
                + d.name
                + """</td>
            <td>"""
                + d.description[0:500]
                + """</td>
            <td>"""
                + assets
                + """</td>
            <td>"""
                + d.id
                + """</td>
          </tr>"""
            )

        return initial_boilerplate + rows + end_boilerplate


class Dataset:
    def __init__(
        self,
        raw: Any,
        client: AbstractNodeClient,
        description: str,
        name: str,
        id: UID,
        key: int,
        data: Any,
        tags: Optional[List[str]] = None,
    ) -> None:
        self.raw = raw
        self.description = description
        self.name = name
        self.id = id
        self.tags = tags if tags is not None else []
        self.data = data
        self.client = client
        self.key = key

    @property
    def pandas(self) -> pd.DataFrame:
        return pd.DataFrame(self.raw)

    @property
    def assets(self) -> Any:
        return self.data

    def __getitem__(self, key: str) -> Any:
        keys = list()
        for d in self.data:
            if d["name"] == key:
                return self.client.store.get(d["id"])  # type: ignore
            keys.append(d["name"])

        raise KeyError(
            "Asset '" + key + "' doesn't exist! Try one of these: " + str(keys)
        )

    def _repr_html_(self) -> str:

        print("Dataset: " + self.name)
        print("Description: " + self.description)
        print()

        rows = ""

        data = self.data

        if len(data) > 15:
            print(
                "WARNING: Too many assets to print... truncating... You "
                "may run \n\n assets = my_dataset.assets \n\nto view receive a "
                "dictionary you can parse through using Python\n(as opposed to blowing up your notebook"
                " with a massive printed table).\n"
            )
            data = data[0:15]

        assets = ""
        for _, a in enumerate(data):
            assets += '["' + a["name"] + '"] -> ' + a["dtype"] + "<br /><br />"

            rows += (
                """

              <tr>
            <td>[\""""
                + a["name"]
                + """\"]</td>
            <td>"""
                + a["dtype"]
                + """</td>
            <td>"""
                + a["shape"]
                + """</td>
          </tr>"""
            )
        end_boilerplate = """
        </table>

        """

        return initial_boilerplate + rows + end_boilerplate

    def add(self, name: str, value: Any, skip_checks: bool = False) -> None:
        """Add a new asset to the dataset.

        Args:
            name (str): Name of the asset
            value (dict): Value of the asset
        """

        # relative
        from .....lib.python.util import downcast

        if not skip_checks:
            if not isinstance(value, Tensor) or not isinstance(
                getattr(value, "child", None), ADPTensor
            ):
                raise Exception(
                    "ERROR: all private assets must be NumPy ndarray.int32 assets "
                    + "with proper Differential Privacy metadata applied.\n"
                    + "\n"
                    + "Example: syft.Tensor(np.ndarray([1,2,3,4]).astype(np.int32)).private()\n\n"
                    + "and then follow the wizard. 🧙"
                )
                # print(
                #     "\n\nWARNING - Non-DP Asset: You just passed in a asset '"
                #     + name
                #     + "' which cannot be tracked with differential privacy because it is a "
                #     + str(type(value))
                #     + " object.\n\n"
                #     + "This means you'll need to manually approve any requests which "
                #     + "leverage this data. If this is ok with you, proceed. If you'd like to use "
                #     + "automatic differential privacy budgeting, please pass in a DP-compatible tensor type "
                #     + "such as by calling .private() on a sy.Tensor with a np.int32 or np.float32 inside."
                # )
                #
                # pref = input("Are you sure you want to proceed? (y/n)")
                #
                # while pref != "y" and pref != "n":
                #     pref = input(
                #         "Invalid input '" + pref + "', please specify 'y' or 'n'."
                #     )
                # if pref == "n":
                #     raise Exception("Dataset loading cancelled.")

            existing_asset_names = [d.get("name") for d in self.data]
            if name in existing_asset_names:
                raise KeyError(
                    f"Asset with name: `{name}` already exists. "
                    "Please use a different name."
                )

            asset = {name: value}
            asset = downcast(asset)
            binary_dataset = serialize(asset, to_bytes=True)

            metadata = {"dataset_id": bytes(str(self.id), "utf-8")}
            metadata = downcast(metadata)

            sys.stdout.write("\rLoading dataset... uploading...")
            # Add a new asset to the dataset pointer
            DatasetRequestAPI(self.client).create_syft(
                dataset=binary_dataset, metadata=metadata, platform="syft"
            )
            sys.stdout.write("\rLoading dataset... uploading... \nSUCCESS!")
            self.refresh()

    def delete(self, name: str, skip_check: bool = False) -> bool:
        """Delete the asset with the given name."""

        asset_id = None

        for d in self.data:
            if d["name"] == name:
                asset_id = d["id"]  # Id of the first matching name
                break

        if asset_id is None:
            raise KeyError(f"The asset with name `{name}` does not exists.")

        if not skip_check:
            pref = input(
                f"You are about to permanantely delete the asset `{name}` ? 🚨 \n"
                "Please enter y/n to proceed: "
            )
            while pref != "y" and pref != "n":
                pref = input("Invalid input '" + pref + "', please specify 'y' or 'n'.")
            if pref == "n":
                sys.stdout.write("Asset deletion cancelled.")
                return False

        DatasetRequestAPI(self.client).delete(
            dataset_id=self.id, bin_object_id=asset_id
        )
        self.refresh()

        return True

    def refresh(self) -> None:
        """Update data to its latest state."""

        datasets = DatasetRequestAPI(self.client).all()
        self.data = datasets[self.key].get("data", [])

    def iter(self, exclude: Optional[List[str]] = None) -> Iterable:
        """Generate an asset iterable."""

        exclude = [] if exclude is None else exclude

        for asset in self.data:
            asset_name = asset["name"]
            if asset_name not in exclude:
                asset_id = asset["id"].replace("-", "")
                yield self.client.store.get(asset_id)  # type: ignore



"""Beaver Triples Protocol.

D. Beaver. *Efficient multiparty protocols using circuit randomization*.
In J. Feigenbaum, editor, CRYPTO, volume **576** of Lecture Notes in
Computer Science, pages 420–432. Springer, 1991.
"""


# stdlib
from copy import deepcopy
import secrets
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

# third party
import numpy as np

# relative
from ....tensor.config import DEFAULT_RING_SIZE
from ....tensor.smpc.mpc_tensor import MPCTensor
from ....tensor.smpc.share_tensor import ShareTensor
from ....tensor.smpc.utils import RING_SIZE_TO_TYPE
from ....tensor.smpc.utils import count_wraps
from ...store import register_primitive_generator
from ...store import register_primitive_store_add
from ...store import register_primitive_store_get
from ...store.exceptions import EmptyPrimitiveStore

ttp_generator = np.random.default_rng()


def get_child(lst: List[Any]) -> List:
    return [value.child for value in lst]


def _get_triples(
    op_str: str,
    nr_parties: int,
    parties_info: List[Any],
    a_shape: Tuple[int],
    b_shape: Tuple[int],
    nr_instances: int = 1,
    ring_size: int = DEFAULT_RING_SIZE,
    **kwargs: Dict[Any, Any],
) -> Tuple[Tuple[Tuple[ShareTensor, ShareTensor, ShareTensor]]]:
    """Get triples.

    The Trusted Third Party (TTP) or Crypto Provider should provide this triples Currently,
    the one that orchestrates the communication provides those triples.".

    Args:
        op_str (str): Operator string.
        nr_parties (int): Number of parties
        parties_info (List[Any]): Parties connection information.
        a_shape (Tuple[int]): Shape of a from beaver triples protocol.
        b_shape (Tuple[int]): Shape of b part from beaver triples protocol.
        ring_size (int) : Ring Size of the triples to generate.
        kwargs: Arbitrary keyword arguments for commands.

    Returns:
        List[List[3 x List[ShareTensor, ShareTensor, ShareTensor]]]:
        The generated triples a,b,c for each party.

    Raises:
        ValueError: If the triples are not consistent.
        ValueError: If the share class is invalid.
    """
    # relative
    from ..... import Tensor

    numpy_type = RING_SIZE_TO_TYPE[ring_size]
    cmd = ShareTensor.get_op(ring_size, op_str)

    min_value, max_value = ShareTensor.compute_min_max_from_ring(ring_size)

    triples = []
    for _ in range(nr_instances):
        seed_przs = secrets.randbits(32)
        a_rand = Tensor(
            ttp_generator.integers(
                low=min_value,
                high=max_value,
                size=a_shape,
                endpoint=True,
                dtype=numpy_type,
            )
        )

        a_shares = MPCTensor._get_shares_from_local_secret(
            secret=deepcopy(a_rand),
            parties_info=parties_info,  # type: ignore
            shape=a_shape,
            seed_przs=seed_przs,
            ring_size=ring_size,
        )
        a_shares = get_child(a_shares)

        seed_przs = secrets.randbits(32)
        b_rand = Tensor(
            ttp_generator.integers(
                low=min_value,
                high=max_value,
                size=b_shape,
                endpoint=True,
                dtype=numpy_type,
            )
        )

        b_shares = MPCTensor._get_shares_from_local_secret(
            secret=deepcopy(b_rand),
            parties_info=parties_info,  # type: ignore
            shape=b_shape,
            seed_przs=seed_przs,
            ring_size=ring_size,
        )
        b_shares = get_child(b_shares)
        seed_przs = secrets.randbits(32)
        # TODO: bitwise and on passthrough tensor raises exception
        # hence we do it on numpy array itself.
        c_val = Tensor(cmd(a_rand.child, b_rand.child))
        c_shares = MPCTensor._get_shares_from_local_secret(
            secret=deepcopy(c_val),
            parties_info=parties_info,  # type: ignore
            shape=c_val.shape,  # type: ignore
            seed_przs=seed_przs,
            ring_size=ring_size,
        )
        c_shares = get_child(c_shares)
        # We are always creating an instance
        triples.append((a_shares, b_shares, c_shares))

    """
    Example -- for n_instances=2 and n_parties=2:
    For Beaver Triples the "res" would look like:
    res = [
        ([a0_sh_p0, a0_sh_p1], [b0_sh_p0, b0_sh_p1], [c0_sh_p0, c0_sh_p1]),
        ([a1_sh_p0, a1_sh_p1], [b1_sh_p0, b1_sh_p1], [c1_sh_p0, c1_sh_p1])
    ]

    We want to send to each party the values they should hold:
    primitives = [
        [[a0_sh_p0, b0_sh_p0, c0_sh_p0], [a1_sh_p0, b1_sh_p0, c1_sh_p0]], # (Row 0)
        [[a0_sh_p1, b0_sh_p1, c0_sh_p1], [a1_sh_p1, b1_sh_p1, c1_sh_p1]]  # (Row 1)
    ]

    The first party (party 0) receives Row 0 and the second party (party 1) receives Row 1
    """

    res_triples = list(zip(*[zip(*triple) for triple in triples]))

    return res_triples  # type: ignore


# Beaver Operations defined for Multiplication


@register_primitive_generator("beaver_mul")
def get_triples_mul(
    *args: List[Any], **kwargs: Dict[Any, Any]
) -> Tuple[Tuple[Tuple[ShareTensor, ShareTensor, ShareTensor]]]:
    """Get the beaver triples for the multiplication operation.

    Args:
        *args (List[ShareTensor]): Named arguments of :func:`beaver.__get_triples`.
        **kwargs (List[ShareTensor]): Keyword arguments of :func:`beaver.__get_triples`.

    Returns:
        Tuple[Tuple[ShareTensor, ShareTensor, ShareTensor]]: The generated triples a,b,c
        for the mul operation.
    """
    return _get_triples("mul", *args, **kwargs)  # type: ignore


@register_primitive_store_add("beaver_mul")
def mul_store_add(
    store: Dict[str, List[Any]],
    primitives: List[Any],
    a_shape: Tuple[int],
    b_shape: Tuple[int],
    ring_size: Union[int, str],
) -> None:
    """Add the primitives required for the "mul" operation to the CryptoStore.

    Arguments:
        store (Dict[str, List[Any]]): the CryptoStore
        primitives (List[Any]): the list of primitives
        a_shape (Tuple[int]): the shape of the first operand
        b_shape (Tuple[int]): the shape of the second operand
    """
    ring_size = int(ring_size)
    config_key = f"beaver_mul_{a_shape}_{b_shape}_{ring_size}"
    if config_key in store:
        store[config_key].extend(list(primitives))
    else:
        store[config_key] = list(primitives)


@register_primitive_store_get("beaver_mul")
def mul_store_get(
    store: Dict[str, List[Any]],
    a_shape: Tuple[int, ...],
    b_shape: Tuple[int, ...],
    ring_size: int,
    remove: bool = True,
) -> Any:
    """Retrieve the primitives from the CryptoStore.

    Those are needed for executing the "mul" operation.

    Args:
        store (Dict[str, List[Any]]): The CryptoStore.
        a_shape (Tuple[int]): The shape of the first operand.
        b_shape (Tuple[int]): The shape of the second operand.
        remove (bool): True if the primitives should be removed from the store.

    Returns:
        Any: The primitives required for the "mul" operation.

    Raises:
        EmptyPrimitiveStore: If no primitive in the store for config_key.
    """
    config_key = f"beaver_mul_{tuple(a_shape)}_{tuple(b_shape)}_{ring_size}"

    try:
        primitives = store[config_key]
    except KeyError:
        raise EmptyPrimitiveStore(f"{config_key} does not exists in the store")

    try:
        primitive = primitives[0]
    except Exception:
        raise EmptyPrimitiveStore(f"No primitive in the store for {config_key}")

    if remove:
        del primitives[0]

    return primitive


# Beaver Operations defined for Matrix Multiplication


@register_primitive_generator("beaver_matmul")
def get_triples_matmul(
    *args: List[Any], **kwargs: Dict[Any, Any]
) -> Tuple[List[ShareTensor], List[ShareTensor], List[ShareTensor]]:
    """Get the beaver triples for the matmul  operation.

    Args:
        *args (List[ShareTensor]): Named arguments of :func:`beaver.__get_triples`.
        **kwargs (List[ShareTensor]): Keyword arguments of :func:`beaver.__get_triples`.

    Returns:
        Tuple[Tuple[ShareTensor, ShareTensor, ShareTensor]]: The generated triples a,b,c
        for the matmul operation.
    """
    return _get_triples("matmul", *args, **kwargs)  # type: ignore


@register_primitive_store_add("beaver_matmul")
def matmul_store_add(
    store: Dict[str, List[Any]],
    primitives: List[Any],
    a_shape: Tuple[int],
    b_shape: Tuple[int],
    ring_size: Union[int, str],
) -> None:
    """Add the primitives required for the "matmul" operation to the CryptoStore.

    Args:
        store (Dict[str, List[Any]]): The CryptoStore.
        primitives (List[Any]): The list of primitives
        a_shape (Tuple[int]): The shape of the first operand.
        b_shape (Tuple[int]): The shape of the second operand.

    """
    ring_size = int(ring_size)
    config_key = f"beaver_matmul_{a_shape}_{b_shape}_{ring_size}"
    if config_key in store:
        store[config_key].extend(list(primitives))
    else:
        store[config_key] = list(primitives)


@register_primitive_store_get("beaver_matmul")
def matmul_store_get(
    store: Dict[str, List[Any]],
    a_shape: Tuple[int, ...],
    b_shape: Tuple[int, ...],
    ring_size: int,
    remove: bool = True,
) -> Any:
    """Retrieve the primitives from the CryptoStore.

    Those are needed for executing the "matmul" operation.

    Args:
        store (Dict[str, List[Any]]): The CryptoStore.
        a_shape (Tuple[int]): The shape of the first operand.
        b_shape (Tuple[int]): The shape of the second operand.
        remove (bool): True if the primitives should be removed from the store.

    Returns:
        Any: The primitives required for the "matmul" operation.

    Raises:
        EmptyPrimitiveStore: If no primitive in the store for config_key.
    """
    config_key = f"beaver_matmul_{tuple(a_shape)}_{tuple(b_shape)}_{ring_size}"

    try:
        primitives = store[config_key]
    except KeyError:
        raise EmptyPrimitiveStore(f"{config_key} does not exists in the store")

    try:
        primitive = primitives[0]
    except Exception:
        raise EmptyPrimitiveStore(f"No primitive in the store for {config_key}")

    if remove:
        del primitives[0]

    return primitive


# Trusted Third Party (TTP) Operations for Public Division
# Code Adapted from Crypten Project:  https://github.com/facebookresearch/CrypTen


@register_primitive_generator("beaver_wraps")
def count_wraps_rand(
    nr_parties: int,
    shape: Tuple[int],
    parties_info: List[Any],
    nr_instances: int = 1,
    ring_size: int = DEFAULT_RING_SIZE,
) -> List[Tuple[Tuple[ShareTensor, ShareTensor]]]:
    """Count wraps random.
    The Trusted Third Party (TTP) or Crypto provider should generate:
    - a set of shares for a random number
    - a set of shares for the number of wraparounds for that number
    Those shares are used when doing a public division, such that the
    end result would be the correct one.

    Args:
        nr_parties (int): Number of parties
        shape (Tuple[int]): The shape for the random value
        parties_info (List[Any]): Parties connection information.
        shape (Tuple[int]): the shape of the numerator
        ring_size (int) : Ring Size of the operation.
        kwargs: Arbitrary keyword arguments for commands.

    Returns:
        List[List[List[ShareTensor, ShareTensor]]: a list of instaces with the shares
        for a random integer value and shares for the number of wraparounds that are done when
        reconstructing the random value
    """
    # relative
    from ..... import Tensor

    numpy_type = RING_SIZE_TO_TYPE[ring_size]
    min_value, max_value = ShareTensor.compute_min_max_from_ring(ring_size)

    primitives = []

    for _ in range(nr_instances):

        seed_przs = secrets.randbits(32)
        rand_val = Tensor(
            ttp_generator.integers(
                low=min_value,
                high=max_value,
                size=shape,
                endpoint=True,
                dtype=numpy_type,
            )
        )

        r_shares = MPCTensor._get_shares_from_local_secret(
            secret=deepcopy(rand_val),
            parties_info=parties_info,  # type: ignore
            shape=shape,
            seed_przs=seed_przs,
            ring_size=ring_size,
        )
        r_shares = get_child(r_shares)

        seed_przs = secrets.randbits(32)
        wraps = Tensor(count_wraps([share.child for share in r_shares]))

        theta_r_shares = MPCTensor._get_shares_from_local_secret(
            secret=deepcopy(wraps),
            parties_info=parties_info,  # type: ignore
            shape=shape,
            seed_przs=seed_przs,
            ring_size=ring_size,
        )
        theta_r_shares = get_child(theta_r_shares)

        # For now We are always creating only an instance
        primitives.append((r_shares, theta_r_shares))

    res_primitives = list(zip(*[zip(*primitive) for primitive in primitives]))

    return res_primitives  # type: ignore


@register_primitive_store_add("beaver_wraps")
def wraps_store_add(
    store: Dict[str, List[Any]],
    primitives: List[Any],
    shape: Tuple[int],
    ring_size: Union[int, str],
) -> None:
    """Add the primitives required for the public division operation to the CryptoStore.

    Arguments:
        store (Dict[str, List[Any]]): the CryptoStore
        primitives (List[Any]): the list of primitives
        shape (Tuple[int]): the shape of the numerator
        ring_size (int): Ring size of the operation.
    """
    ring_size = int(ring_size)
    config_key = f"beaver_wraps_{shape}_{ring_size}"
    if config_key in store:
        store[config_key].extend(list(primitives))
    else:
        store[config_key] = list(primitives)


@register_primitive_store_get("beaver_wraps")
def wraps_store_get(
    store: Dict[str, List[Any]],
    shape: Tuple[int, ...],
    ring_size: int,
    remove: bool = True,
) -> Any:
    """Retrieve the primitives from the CryptoStore.

    Those are needed for executing the public division operation.

    Args:
        store (Dict[str, List[Any]]): The CryptoStore.
        shape (Tuple[int]): the shape of the numerator
        ring_size (int): Ring size of the operation.
        remove (bool): True if the primitives should be removed from the store.

    Returns:
        Any: The primitives required for the public division operation.

    Raises:
        EmptyPrimitiveStore: If no primitive in the store for config_key.
    """
    config_key = f"beaver_wraps_{tuple(shape)}_{ring_size}"

    try:
        primitives = store[config_key]
    except KeyError:
        raise EmptyPrimitiveStore(f"{config_key} does not exists in the store")

    try:
        primitive = primitives[0]
    except Exception:
        raise EmptyPrimitiveStore(f"No primitive in the store for {config_key}")

    if remove:
        del primitives[0]

    return primitive



# stdlib
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ......grid import GridURL
from .....common.serde.serializable import serializable
from ....domain_msg_registry import DomainMessageRegistry
from ....network_msg_registry import NetworkMessageRegistry
from ....node_service import NodeServiceInterface
from ...permissions.user_permissions import NoRestriction
from ...permissions.user_permissions import UserIsOwner
from ..generic_payload.syft_message import NewSyftMessage as SyftMessage
from ..generic_payload.syft_message import ReplyPayload
from ..generic_payload.syft_message import RequestPayload
from .node_credentials import NodeCredentials


# ExchangeCredentials Messages
# Step 1: InitiateExchangeCredentialsWithNodeMessage
# Step 2: ExchangeCredentialsWithNodeMessage
@serializable(recursive_serde=True)
class InitiateExchangeCredentialsWithNodeMessage(
    SyftMessage, DomainMessageRegistry, NetworkMessageRegistry
):
    permissions = [NoRestriction | UserIsOwner]  # UserIsOwner not working

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a Request."""

        target_node_url: str

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a Response."""

        message: str = "Node credentials exchanged."

    request_payload_type = Request
    reply_payload_type = Reply

    def run(  # type: ignore
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore
        """Validates the request parameters and sends a ExchangeCredentialsWithNodeMessage.

        Args:
            node (NodeServiceInterface): Node either Domain or Network.
            verify_key (Optional[VerifyKey], optional): User signed verification key. Defaults to None.

        Raises:
            InvalidNodeCredentials: If the credentials are invalid

        Returns:
            ReplyPayload: Message on successful exchange.
        """

        # TODO: get the client from the node and use a hashmap of uuid and / or url
        # to cache existing client objects to target nodes
        # get client for target node

        # relative
        from ......grid.client.client import connect

        # we may be trying to call another local test host like localhost so we need
        # to also call as_container_host
        target_url = (
            GridURL.from_url(self.payload.target_node_url)
            .with_path("/api/v1")
            .as_container_host(container_host=node.settings.CONTAINER_HOST)
        )

        # we use our local keys so that signing and verification matches our node
        target_client = connect(url=target_url, timeout=10, user_key=node.signing_key)

        # send credentials to other node
        credentials = node.get_credentials()

        # see ExchangeCredentialsWithNodeMessage below
        signed_response = target_client.networking.exchange_credentials_with_node(
            credentials=credentials
        )

        # since we're getting back the SignedMessage it can't hurt to check once more
        if not signed_response.is_valid:
            raise Exception(
                "Response was signed by a fake key or was corrupted in transit."
            )

        response_credentials = NodeCredentials(
            **signed_response.message.payload.credentials  # type: ignore
        )

        # can we get the associated verify_key?
        response_credentials.validate(key=signed_response.verify_key)

        # add response NodeCredentials
        node.node.add_or_update_node_credentials(credentials=response_credentials)

        return self.Reply()


@serializable(recursive_serde=True)
class ExchangeCredentialsWithNodeMessage(
    SyftMessage, DomainMessageRegistry, NetworkMessageRegistry
):
    permissions = [NoRestriction]

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a Request."""

        credentials: dict

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a Response."""

        credentials: dict
        message: str = "Node credentials validated and added."

    request_payload_type = Request
    reply_payload_type = Reply

    def run(  # type: ignore
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore
        """Validates the request parameters and exchanges credentials.

        Args:
            node (NodeServiceInterface): Node either Domain or Network.
            verify_key (Optional[VerifyKey], optional): User signed verification key. Defaults to None.

        Raises:
            InvalidNodeCredentials: If the credentials are invalid

        Returns:
            ReplyPayload: Message on successful user creation.
        """

        request_credentials = NodeCredentials(**self.payload.credentials)

        # check the key we will store and compare to the node table is from the holder
        # of the private key who signed the message
        request_credentials.validate(key=verify_key)

        # validate NodeCredentials
        node.node.add_or_update_node_credentials(credentials=request_credentials)

        # respond with this nodes NodeCredentials
        response_credentials = node.get_credentials()

        return self.Reply(**{"credentials": response_credentials})



# stdlib
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np
from numpy.typing import NDArray

# relative
from .. import activations
from ....common.serde.serializable import serializable
from ...autodp.phi_tensor import GammaTensor
from ...autodp.phi_tensor import PhiTensor
from .base import Layer


@serializable(recursive_serde=True)
class BatchNorm(Layer):
    __attr_allowlist__ = (
        "epsilon",
        "momentum",
        "axis",
        "activation",
        "beta",
        "dbeta",
        "gamma",
        "dgamma",
        "cache",
        "input_shape",
        "out_shape",
    )

    def __init__(
        self,
        epsilon: float = 1e-6,
        momentum: float = 0.9,
        axis: int = 0,
        activation: Optional[str] = None,
    ):
        self.epsilon = epsilon
        self.momentum = momentum
        self.axis = axis
        self.activation = activations.get(activation)

        self.beta: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.dbeta: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.gamma: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.dgamma: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.cache: Tuple = tuple()
        self.input_shape: Optional[Tuple[int, ...]] = None
        self.out_shape: Optional[Tuple[int, ...]] = None

    def connect_to(self, prev_layer: Layer) -> None:

        # N, C, H, W -> out_shape
        _, C, _, _ = prev_layer.out_shape  # type: ignore
        self.beta = np.zeros((1, C, 1, 1))
        self.gamma = np.zeros((1, C, 1, 1))
        self.input_shape = prev_layer.out_shape  # type: ignore
        self.out_shape = prev_layer.out_shape  # type: ignore

    def forward(
        self, input: Union[PhiTensor, GammaTensor], *args: Any, **kwargs: Any
    ) -> Union[PhiTensor, GammaTensor]:
        # N, D = x.shape

        # step1: calculate the mean
        # mean calc along N, H, W axis
        # Reference: https://arxiv.org/pdf/1803.08494.pdf
        mean = input.mean(axis=(0, 2, 3), keepdims=True)

        xmu = input - mean

        # step2:
        # sq = xmu ** 2
        # var = 1. / N * np.sum(sq, axis=(N, H, W))
        # std calc along N, H, W axis
        # Reference: https://arxiv.org/pdf/1803.08494.pdf
        var = xmu.std(axis=(0, 2, 3), keepdims=True)

        # step3:
        sqrtvar = (var + self.epsilon).sqrt()

        ivar = 1.0 / sqrtvar

        # step4: normalization->x^
        xhat = xmu * ivar

        # step5: scale and shift
        gammax = xhat * self.gamma

        out = gammax + self.beta

        out = self.activation.forward(out) if self.activation is not None else out

        self.cache = (xhat, xmu, ivar, sqrtvar, var)
        return out

    def backward(
        self,
        pre_grad: Union[PhiTensor, GammaTensor],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> Union[PhiTensor, GammaTensor]:
        """
        If you get stuck, here's a resource:
        https://kratzert.github.io/2016/02/12/understanding-the-
        gradient-flow-through-the-batch-normalization-layer.html

        # mean, std, sum is calculated along N, H, W axis
        # Reference: https://arxiv.org/pdf/1803.08494.pdf

        Note:
            - I removed the np.ones() at a few places where I
               thought it wasn't making a difference
            - I occasionally have kernel crashes on my 8GB machine
            when running this. Perhaps too many large temp vars?
            could also be due to too many large numbers.
        """

        xhat, xmu, ivar, sqrtvar, var = self.cache

        # pre_grad.shape -> N, C, H, W
        N, _, _, _ = pre_grad.shape

        pre_grad = (
            self.activation.derivative(pre_grad)
            if self.activation is not None
            else pre_grad
        )

        # step6
        # sum is calculated along N, H, W axis
        self.dbeta = pre_grad.sum(axis=(0, 2, 3), keepdims=True)
        dgammax = pre_grad

        # sum is calculated along N, H, W axis
        self.dgamma = (dgammax * xhat).sum(axis=(0, 2, 3), keepdims=True)
        dxhat = dgammax * self.gamma

        # step5
        # var is calculated along N, H, W axis
        divar = (dxhat * xmu).sum(axis=(0, 2, 3), keepdims=True)
        dxmu1 = dxhat * ivar

        # step4
        dsqrtvar = -1.0 / (sqrtvar * sqrtvar) * divar
        inv_var_eps_sqrt = 1.0 / (var + self.epsilon).sqrt()
        dvar = dsqrtvar * 0.5 * inv_var_eps_sqrt

        # step3
        dsq = (pre_grad.ones_like()) * dvar * (1.0 / N)
        dxmu2 = xmu * dsq * 2

        # step2
        dx1 = dxmu1 + dxmu2

        # step1
        # sum is calculated along N, H, W axis
        dmu = (dxmu1 + dxmu2).sum(axis=(0, 2, 3), keepdims=True) * -1
        dx2 = (pre_grad.ones_like()) * dmu * (1 / N)

        # step0 done!
        dx = dx1 + dx2

        return dx

    @property
    def params(self) -> Tuple[Union[PhiTensor, GammaTensor, NDArray, None], ...]:
        return self.beta, self.gamma

    @params.setter
    def params(
        self, new_params: List[Union[np.ndarray, PhiTensor, GammaTensor]]
    ) -> None:
        if len(new_params) != 2:
            raise ValueError(
                f"Expected Two values Update Params has length{len(new_params)}"
            )
        self.beta, self.gamma = new_params

    @property
    def grads(self) -> Tuple[Union[PhiTensor, GammaTensor, NDArray, None], ...]:
        return self.dbeta, self.dgamma



# stdlib
from typing import Optional

# relative
from .....core.common.message import SignedMessage
from ...abstract.node import AbstractNodeClient
from ..action.exception_action import ExceptionMessage
from ..node_service.node_credential.node_credential_messages import (
    ExchangeCredentialsWithNodeMessage,
)
from ..node_service.node_credential.node_credential_messages import (
    InitiateExchangeCredentialsWithNodeMessage,
)
from ..node_service.node_credential.node_credential_messages import NodeCredentials
from ..node_service.node_route.node_route_messages import (
    InitiateRouteUpdateToNodeMessage,
)
from ..node_service.node_route.node_route_messages import (
    NotifyNodeWithRouteUpdateMessage,
)
from ..node_service.node_route.node_route_messages import RoutesListMessage
from ..node_service.node_route.node_route_messages import VerifyRouteUpdateMessage
from ..node_service.node_route.route_update import RouteUpdate
from .new_request_api import ClientLike
from .new_request_api import NewRequestAPI


class NodeNetworkingAPI(NewRequestAPI):
    def __init__(self, client: AbstractNodeClient):
        self.client = client

    def initiate_exchange_credentials(
        self, client: ClientLike
    ) -> InitiateExchangeCredentialsWithNodeMessage.Reply:
        # this should be run by a user on their node (probably a domain)
        target_url = self.get_client_url(client).base_url
        signed_msg = InitiateExchangeCredentialsWithNodeMessage(
            address=self.client.address,
            reply_to=self.client.address,
            kwargs={"target_node_url": str(target_url)},
        ).sign(
            signing_key=self.client.signing_key
        )  # type: ignore

        response = self.client.send_immediate_msg_with_reply(msg=signed_msg)
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            return response

    def exchange_credentials_with_node(
        self, credentials: NodeCredentials
    ) -> SignedMessage:
        # this should be run by a node (probably a domain) against a network
        # in this case the self context will be a NetworkClient

        signed_msg = ExchangeCredentialsWithNodeMessage(
            address=self.client.address,
            reply_to=self.client.address,
            kwargs={"credentials": {**credentials}},
        ).sign(
            signing_key=self.client.signing_key
        )  # type: ignore

        signed_response = self.client.send_immediate_msg_with_reply(
            msg=signed_msg, return_signed=True
        )

        if isinstance(signed_response, ExceptionMessage):
            raise signed_response.exception_type
        else:
            return signed_response

    def add_route_for(
        self,
        client: ClientLike,
        source_node_url: Optional[str] = None,
        private: bool = False,
        autodetect: bool = True,
    ) -> InitiateRouteUpdateToNodeMessage.Reply:
        # this should be run by a user on their node (probably a domain)
        target_url = self.get_client_url(client).base_url

        route_update = RouteUpdate(
            source_node_uid=self.client.id.no_dash,
            source_node_url=source_node_url,
            private=private,
            autodetect=autodetect,
        )

        route_update.validate()

        signed_msg = InitiateRouteUpdateToNodeMessage(
            address=self.client.address,
            reply_to=self.client.address,
            kwargs={
                "target_node_url": str(target_url),
                "route_update": {**route_update},
            },
        ).sign(
            signing_key=self.client.signing_key
        )  # type: ignore

        response = self.client.send_immediate_msg_with_reply(msg=signed_msg)
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            return response

    def notify_node_with_route_update(
        self, route_update: RouteUpdate
    ) -> NotifyNodeWithRouteUpdateMessage.Reply:
        # this should be run by a node (probably a domain) against a network
        # in this case the self context will be a NetworkClient
        signed_msg = NotifyNodeWithRouteUpdateMessage(
            address=self.client.address,
            reply_to=self.client.address,
            kwargs={"route_update": {**route_update}},
        ).sign(
            signing_key=self.client.signing_key
        )  # type: ignore

        response = self.client.send_immediate_msg_with_reply(msg=signed_msg)
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            return response

    def verify_route(
        self,
    ) -> SignedMessage:
        # this should be run by a node (probably a domain) responding to the
        # previous request to make sure the route is reachable
        # we return the SignedMessage so we can double check the Domain we reached
        # is the one we are hoping to reach
        signed_msg = VerifyRouteUpdateMessage(
            address=self.client.address,
            reply_to=self.client.address,
            kwargs={},
        ).sign(
            signing_key=self.client.signing_key
        )  # type: ignore

        signed_response = self.client.send_immediate_msg_with_reply(
            msg=signed_msg, return_signed=True
        )

        if isinstance(signed_response, ExceptionMessage):
            raise signed_response.exception_type
        else:
            return signed_response

    def list_routes(
        self, client: ClientLike, timeout: Optional[int] = None
    ) -> RoutesListMessage.Reply:
        # this should be run by a Domain owner against a network
        target_client = self.get_client(client)
        signed_msg = RoutesListMessage(
            address=target_client.address,
            reply_to=target_client.address,
            kwargs={},
        ).sign(
            signing_key=target_client.signing_key
        )  # type: ignore

        response = target_client.send_immediate_msg_with_reply(
            msg=signed_msg, timeout=timeout
        )
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            return response.payload



# stdlib
from typing import Any
from typing import Dict
from typing import List

# third party
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Query

# relative
from ..exceptions import RoleNotFoundError
from ..node_table.roles import Role
from .database_manager import DatabaseManager


class RoleManager(DatabaseManager):
    schema = Role

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=RoleManager.schema, db=database)

    @property
    def ds_role(self) -> Role:
        return self.first(name="Data Scientist")

    @property
    def owner_role(self) -> Role:
        return self.first(name="Owner")

    @property
    def compliance_officer_role(self) -> Role:
        return self.first(name="Compliance Officer")

    @property
    def admin_role(self) -> Role:
        return self.first(name="Administrator")

    def _common_roles(self) -> Query:
        return self.db.session.query(self._schema).filter_by(
            can_make_data_requests=True,
            can_triage_data_requests=False,
            can_manage_privacy_budget=False,
            can_create_users=False,
            can_manage_users=False,
            can_edit_roles=False,
            can_manage_infrastructure=False,
            can_upload_data=False,
            can_upload_legal_document=False,
            can_edit_domain_settings=False,
        )

    def can_create_users(self, role_id: int) -> bool:
        role = super().first(id=role_id)
        if role:
            return role.can_create_users
        else:
            return False

    @property
    def common_roles(self) -> List[Role]:
        return self._common_roles().all()

    @property
    def org_roles(self) -> List[Role]:
        return self.db.session.query(self._schema).except_(self._common_roles).all()

    def first(self, **kwargs: Any) -> Role:
        result = super().first(**kwargs)
        if not result:
            raise RoleNotFoundError
        return result

    def query(self, **kwargs: Any) -> Query:
        results = super().query(**kwargs)
        if len(results) == 0:
            raise RoleNotFoundError
        return results

    def set(self, role_id: int, params: Dict[Any, Any]) -> None:
        if self.contain(id=role_id):
            self.modify({"id": role_id}, params)
        else:
            raise RoleNotFoundError



# stdlib
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple
from typing import Union

# third party
import numpy as np
from numpy.typing import NDArray

# relative
from ...adp.data_subject_list import DataSubjectArray
from ..autodp.gamma_tensor import GammaTensor
from ..autodp.phi_tensor import PhiTensor
from ..lazy_repeat_array import lazyrepeatarray


def dp_maximum(
    x: Union[PhiTensor, GammaTensor], y: Union[np.ndarray, PhiTensor, GammaTensor]
) -> Union[PhiTensor, GammaTensor]:
    # TODO: Make this work for GammaTensors
    x_data = x.child
    y_data = y.child if hasattr(y, "child") else y

    output = np.maximum(x_data, y_data)

    min_v, max_v = output.min(), output.max()
    dsl = x.data_subjects  # TODO: fix later
    return PhiTensor(
        child=output,
        data_subjects=dsl,
        min_vals=min_v,
        max_vals=max_v,
    )


def dp_log(input: Union[PhiTensor, GammaTensor]) -> Union[PhiTensor, GammaTensor]:
    if isinstance(input, PhiTensor):
        output = np.log(input.child)
        # min_v, max_v = output.min(), output.max()  # These bounds are a function of private data
        min_v = lazyrepeatarray(
            data=np.log(input.min_vals.data.min()), shape=output.shape
        )
        max_v = lazyrepeatarray(
            data=np.log(input.max_vals.data.max()), shape=output.shape
        )
        dsl = input.data_subjects

        return PhiTensor(
            child=output,
            data_subjects=dsl,
            min_vals=min_v,
            max_vals=max_v,
        )
    elif isinstance(input, GammaTensor):
        output = np.log(input.child)
        # min_v, max_v = output.min(), output.max()  # These bounds are a function of private data
        min_v = lazyrepeatarray(
            data=np.log(input.min_vals.data.min()), shape=output.shape
        )
        max_v = lazyrepeatarray(
            data=np.log(input.max_vals.data.max()), shape=output.shape
        )
        dsl = input.data_subjects

        return GammaTensor(
            child=output,
            data_subjects=dsl,
            min_vals=min_v,
            max_vals=max_v,
        )
    else:
        raise NotImplementedError(f"Undefined behaviour for type: {type(input)}")


def dp_zeros(shape: Tuple) -> Union[PhiTensor, GammaTensor]:
    """
    TODO: Passing in the shape seems unnecessary- it can be inferred from data_subjects.data_subjects_indexed.shape
    output = np.zeros_like(data_subjects.data_subjects_indexed)

    :param shape:
    :param data_subjects:
    :return:
    """
    output = np.zeros(shape)

    # Create Empty DataSubjectSet array
    data_subjects = np.broadcast_to(np.array([DataSubjectArray(set())]), output.shape)

    return GammaTensor(
        child=output,
        data_subjects=data_subjects,
        min_vals=lazyrepeatarray(0, shape),
        max_vals=lazyrepeatarray(1, shape),
    )


def dp_pad(
    input: Union[PhiTensor, GammaTensor],
    width: Union[int, Sequence],
    padding_mode: str = "reflect",
    **kwargs: Dict[Any, Any],
) -> Union[PhiTensor, GammaTensor]:

    data = input.child
    output_data: NDArray = np.pad(data, width, mode=padding_mode, **kwargs)
    min_v = lazyrepeatarray(
        data=min(input.min_vals.data.min(), output_data.min()), shape=output_data.shape
    )
    max_v = lazyrepeatarray(
        data=min(input.max_vals.data.max(), output_data.max()), shape=output_data.shape
    )

    output_data_subjects: NDArray = np.pad(
        input.data_subjects, width, mode=padding_mode, **kwargs
    )

    if isinstance(input, PhiTensor):
        return PhiTensor(
            child=output_data,
            data_subjects=output_data_subjects,
            min_vals=min_v,
            max_vals=max_v,
        )
    elif isinstance(input, GammaTensor):
        return GammaTensor(
            child=output_data,
            data_subjects=output_data_subjects,
            min_vals=min_v,
            max_vals=max_v,
        )
    else:
        raise NotImplementedError(
            f"Padding is not implemented for Input Type: {type(input)}"
        )


def dp_add_at(
    a: Union[PhiTensor, GammaTensor], indices: Tuple, b: Union[PhiTensor, GammaTensor]
) -> Union[PhiTensor, GammaTensor]:
    data_a = a.child
    data_b = b.child

    data_subject_a = a.data_subjects
    # data_subject_b = b.data_subjects

    np.add.at(data_a, indices, data_b)

    # TODO: @Shubham Find np.add.at can be implemented at data_subject_level
    # data_subject_a[indices] = data_subject_a[indices] + data_subject_b
    # The above is an alternate to np.add.at
    # np.add.at(data_subject_a, indices, data_subject_b)

    if isinstance(a, PhiTensor):
        return PhiTensor(
            child=data_a,
            data_subjects=data_subject_a,
            min_vals=data_a.min(),
            max_vals=data_a.max(),
        )
    else:
        return GammaTensor(
            child=data_a,
            data_subjects=data_subject_a,
            min_vals=lazyrepeatarray(data_a.min(), shape=data_a.shape),
            max_vals=lazyrepeatarray(data_a.min(), shape=data_a.shape),
        )


def get_im2col_indices(
    x_shape: Tuple[int, ...],
    field_height: int,
    field_width: int,
    padding: int = 1,
    stride: int = 1,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    # First figure out what the size of the output should be
    N, C, H, W = x_shape
    if (H + 2 * padding - field_height) % stride != 0:
        raise ValueError(
            f"(H + 2 * padding - field_height) H:{H} , padding:{padding} , field_height:{field_height} \
        should be a multiple of stride:{stride}"
        )
    if (W + 2 * padding - field_height) % stride != 0:
        raise ValueError(
            f"(W + 2 * padding - field_height) W:{W} , padding:{padding} , field_height:{field_height} \
        should be a multiple of stride:{stride}"
        )
    out_height = int((H + 2 * padding - field_height) / stride + 1)
    out_width = int((W + 2 * padding - field_width) / stride + 1)

    i0 = np.repeat(np.arange(field_height), field_width)
    i0 = np.tile(i0, C)
    i1 = stride * np.repeat(np.arange(out_height), out_width)
    j0 = np.tile(np.arange(field_width), field_height * C)
    j1 = stride * np.tile(np.arange(out_width), out_height)
    i = i0.reshape(-1, 1) + i1.reshape(1, -1)
    j = j0.reshape(-1, 1) + j1.reshape(1, -1)
    k = np.repeat(np.arange(C), field_height * field_width).reshape(-1, 1)
    return (k.astype(int), i.astype(int), j.astype(int))


def col2im_indices(
    cols: Union[PhiTensor, GammaTensor],
    x_shape: Tuple[int, ...],
    field_height: int = 3,
    field_width: int = 3,
    padding: int = 1,
    stride: int = 1,
) -> Union[PhiTensor, GammaTensor]:
    """An implementation of col2im based on fancy indexing and np.add.at"""
    N, C, H, W = x_shape
    H_padded, W_padded = H + 2 * padding, W + 2 * padding
    x_padded = dp_zeros(shape=(N, C, H_padded, W_padded))

    k, i, j = get_im2col_indices(x_shape, field_height, field_width, padding, stride)
    cols_reshaped = cols.reshape((C * field_height * field_width, -1, N))
    cols_reshaped = cols_reshaped.transpose((2, 0, 1))
    x_padded = dp_add_at(x_padded, (slice(None), k, i, j), cols_reshaped)
    if padding == 0:
        return x_padded
    return x_padded[:, :, padding:-padding, padding:-padding]


def im2col_indices(
    x: Union[PhiTensor, GammaTensor],
    field_height: int,
    field_width: int,
    padding: int = 1,
    stride: int = 1,
) -> Union[PhiTensor, GammaTensor]:
    """An implementation of im2col based on some fancy indexing"""
    # Zero-pad the input
    p = padding

    width: Tuple[Any, ...]
    if len(x.shape) == 4:
        width = ((0, 0), (0, 0), (p, p), (p, p))
    elif len(x.shape) == 3:
        width = ((0, 0), (p, p), (p, p))
    elif len(x.shape) == 2:
        width = ((p, p), (p, p))
    else:
        raise NotImplementedError
    x_padded = dp_pad(x, width)

    k, i, j = get_im2col_indices(x.shape, field_height, field_width, padding, stride)

    cols = x_padded[:, k, i, j]
    C = x.shape[1]
    cols = cols.transpose((1, 2, 0)).reshape((field_height * field_width * C, -1))

    # Not sure why this happens but sometimes this gets a shape of (n, -1)
    if cols.min_vals.shape != cols.shape:
        cols.min_vals.shape = cols.shape
        cols.max_vals.shape = cols.shape
    return cols



# third party
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer

# relative
from . import Base


class UserEnvironment(Base):
    __tablename__ = "userenvironment"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    user = Column(Integer, ForeignKey("syft_user.id"))
    environment = Column(Integer, ForeignKey("environment.id"))

    def __str__(self) -> str:
        return (
            f"<UserGroup id: {self.id}, user: {self.user}, "
            f"group: {self.environment}>"
        )



"""Approximation Functions."""
# relative
from .exp import exp
from .log import log
from .reciprocal import reciprocal
from .utils import sign

APPROXIMATIONS = {
    "log": log,
    "exp": exp,
    "reciprocal": reciprocal,
    "sign": sign,
}

__all__ = ["APPROXIMATIONS"]



# stdlib
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# relative
from .....logger import critical
from .....proto.core.node.common.action.garbage_collect_batched_pb2 import (
    GarbageCollectBatchedAction as GarbageCollectBatchedAction_PB,
)
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address
from ...abstract.node import AbstractNode
from .common import EventualActionWithoutReply


@serializable()
class GarbageCollectBatchedAction(EventualActionWithoutReply):
    def __init__(
        self, ids_at_location: List[UID], address: Address, msg_id: Optional[UID] = None
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.ids_at_location = ids_at_location

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        try:
            for id_at_location in self.ids_at_location:
                node.store.delete(key=id_at_location)
        except Exception as e:
            critical(
                "> GarbageCollectBatchedAction deletion exception "
                + f"{id_at_location} {e}"
            )

    def _object2proto(self) -> GarbageCollectBatchedAction_PB:
        address = sy.serialize(self.address)
        res = GarbageCollectBatchedAction_PB(address=address)
        for id_obj in self.ids_at_location:
            res.ids_at_location.append(sy.serialize(id_obj))

        return res

    @staticmethod
    def _proto2object(
        proto: GarbageCollectBatchedAction_PB,
    ) -> "GarbageCollectBatchedAction":

        ids_at_location = []
        for id_at_location in proto.ids_at_location:
            ids_at_location.append(sy.deserialize(blob=id_at_location))
        addr = sy.deserialize(blob=proto.address)

        return GarbageCollectBatchedAction(
            ids_at_location=ids_at_location,
            address=addr,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return GarbageCollectBatchedAction_PB



# stdlib
from typing import Any
from typing import Optional
from typing import Union

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from typing_extensions import final

# relative
from ...logger import critical
from ..common.message import SignedMessage
from ..common.message import SyftMessage
from ..common.uid import UID
from ..io.location import Location
from ..io.location import SpecificLocation
from .common.node import Node
from .common.node_manager.dict_store import DictStore
from .device_client import DeviceClient
from .vm import VirtualMachine
from .vm_client import VirtualMachineClient


@final
class Device(Node):

    device: SpecificLocation

    client_type = DeviceClient
    child_type = VirtualMachine
    child_type_client_type = VirtualMachineClient

    def __init__(
        self,
        name: Optional[str] = None,
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        device: SpecificLocation = SpecificLocation(),
        vm: Optional[Location] = None,
        device_type: Any = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
        store_type: type = DictStore,
    ):
        super().__init__(
            name=name,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
            store_type=store_type,
        )

        # specific location with name
        self.device = SpecificLocation(name=self.name)

        self.device_type = device_type

        self._register_services()

        self.post_init()

    def post_init(self) -> None:
        Node.set_keys(node=self)
        super().post_init()

    @property
    def icon(self) -> str:
        return "📱"

    @property
    def id(self) -> UID:
        return self.device.id

    def message_is_for_me(self, msg: Union[SyftMessage, SignedMessage]) -> bool:
        # this needs to be defensive by checking device_id NOT device.id or it breaks
        try:
            return msg.address.device_id == self.id and msg.address.vm is None
        except Exception as e:
            critical(f"Error checking if {msg.pprint} is for me on {self.pprint}. {e}")
            return False



# stdlib
from typing import List
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from .....common.group import VERIFYALL
from ....abstract.node import AbstractNode
from ..auth import AuthorizationException
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithoutReply
from .obj_search_permission_messages import ObjectSearchPermissionUpdateMessage


class ImmediateObjectSearchPermissionUpdateService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    @service_auth(root_only=True)
    def process(
        node: AbstractNode,
        msg: ObjectSearchPermissionUpdateMessage,
        verify_key: VerifyKey,
    ) -> None:
        storable_object = node.store.get(msg.target_object_id, proxy_only=True)
        if (
            verify_key != node.root_verify_key
            or verify_key not in storable_object.read_permissions
        ):
            log = (
                f"You do not have permission to update Object with ID: {msg.target_object_id}"
                + "Please submit a request."
            )
            raise AuthorizationException(log)
        target_verify_key = msg.target_verify_key or VERIFYALL
        if msg.add_instead_of_remove:
            storable_object.search_permissions[target_verify_key] = msg.id
        else:
            storable_object.search_permissions.pop(target_verify_key, None)

        node.store[msg.target_object_id] = storable_object

    @staticmethod
    def message_handler_types() -> List[Type[ObjectSearchPermissionUpdateMessage]]:
        return [ObjectSearchPermissionUpdateMessage]



# stdlib
from enum import Enum
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# relative
from ...... import deserialize
from ...... import serialize
from ......logger import critical
from ......logger import debug
from ......logger import traceback
from ......logger import traceback_and_raise
from ......proto.core.node.domain.service.request_message_pb2 import (
    RequestMessage as RequestMessage_PB,
)
from .....common import UID
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.serializable import serializable
from .....io.address import Address
from ...client import Client
from ...node import Node
from ..accept_or_deny_request.accept_or_deny_request_messages import (
    AcceptOrDenyRequestMessage,
)


class RequestStatus(Enum):
    Pending = 1
    Rejected = 2
    Accepted = 3


# TODO: this message conflates Message functionality with Manager/Request_API functionality
# TODO: this needs to be split into two separate pieces of functionality.
@serializable()
class RequestMessage(ImmediateSyftMessageWithoutReply):

    __slots__ = ["name", "request_description", "request_id"]

    def __init__(
        self,
        object_id: UID,
        address: Address,
        requester_verify_key: VerifyKey,
        owner_address: Address,
        status: Optional[str] = "",
        request_type: Optional[str] = "",
        date: Optional[str] = "",
        object_tags: Optional[List[str]] = None,
        object_type: str = "",
        request_description: str = "",
        request_id: Optional[UID] = None,
        owner_client_if_available: Optional[Client] = None,
        destination_node_if_available: Optional[Node] = None,
        timeout_secs: Optional[int] = None,
        requested_budget: Optional[float] = 0.0,
        current_budget: Optional[float] = 0.0,
        user_name: Optional[str] = "",
        user_role: Optional[str] = "",
        user_email: Optional[str] = "",
    ):
        if request_id is None:
            request_id = UID()
        super().__init__(address=address, msg_id=request_id)

        self.object_tags = object_tags if object_tags else []
        self.object_type = object_type
        self.request_description = request_description
        self.request_id = request_id
        self.requester_verify_key = requester_verify_key
        self.object_id = object_id
        self.owner_address = owner_address
        self.owner_client_if_available = owner_client_if_available
        self.destination_node_if_available = destination_node_if_available
        self.timeout_secs = timeout_secs
        self._arrival_time: Optional[float] = None
        self.status: Optional[str] = status
        self.date: Optional[str] = date
        self.request_type: Optional[str] = request_type
        self.user_name: Optional[str] = user_name
        self.user_email: Optional[str] = user_email
        self.user_role: Optional[str] = user_role
        self.requested_budget: float = requested_budget  # type: ignore
        self.current_budget: float = current_budget  # type: ignore

    def accept(self) -> None:
        self.send_msg(accept=True)

    def approve(self) -> None:
        self.accept()

    def grant(self) -> None:
        self.accept()

    def deny(self) -> None:
        self.send_msg(accept=False)

    def send_msg(self, accept: bool) -> None:
        action_name = "Accept" if accept else "Deny"
        if self.owner_client_if_available is not None:
            msg = AcceptOrDenyRequestMessage(
                address=self.owner_client_if_available.address,
                accept=accept,
                request_id=self.id,
            )
            self.owner_client_if_available.send_immediate_msg_without_reply(msg=msg)
        elif self.destination_node_if_available is not None:
            msg = AcceptOrDenyRequestMessage(
                address=self.destination_node_if_available.address,
                accept=accept,
                request_id=self.id,
            )
            try:
                node = self.destination_node_if_available
                router = node.immediate_msg_without_reply_router
                service = router[type(msg)]
                service.process(
                    node=self.destination_node_if_available,
                    msg=msg,
                    verify_key=self.destination_node_if_available.root_verify_key,
                )
            except Exception as e:
                traceback(e)
                critical(f"Tried to {action_name} Message on Node. {e}")
            debug(f"{action_name} Request: " + str(self.id))
        else:
            log = f"No way to dispatch {action_name} Message."
            critical(log)
            traceback_and_raise(Exception(log))

    def reject(self) -> None:
        self.deny()

    def withdraw(self) -> None:
        self.deny()

    @property
    def arrival_time(self) -> Optional[float]:
        return self._arrival_time

    def set_arrival_time(self, arrival_time: float) -> None:
        # used to expire requests as their destination, this should never be serialized
        if self._arrival_time is None:
            self._arrival_time = arrival_time

    def _object2proto(self) -> RequestMessage_PB:
        msg = RequestMessage_PB()
        msg.object_tags.extend(self.object_tags)
        msg.object_type = self.object_type
        msg.status = self.status
        msg.request_type = self.request_type
        msg.date = self.date
        msg.user_name = self.user_name
        msg.user_email = self.user_email
        msg.user_role = self.user_role
        msg.requested_budget = self.requested_budget if self.requested_budget else 0.0
        msg.current_budget = self.current_budget if self.current_budget else 0.0
        msg.request_description = self.request_description
        msg.request_id.CopyFrom(serialize(obj=self.request_id))
        msg.target_address.CopyFrom(serialize(obj=self.address))
        msg.object_id.CopyFrom(serialize(obj=self.object_id))
        msg.owner_address.CopyFrom(serialize(obj=self.owner_address))
        msg.requester_verify_key = bytes(self.requester_verify_key)

        # -1 will represent no timeout, where as 0 is a valid value for timing out
        # immediately after checking if there is a rule in place to accept or deny
        if self.timeout_secs is None or self.timeout_secs < 0:
            self.timeout_secs = -1
        msg.timeout_secs = int(self.timeout_secs)
        return msg

    @staticmethod
    def _proto2object(proto: RequestMessage_PB) -> "RequestMessage":
        request_msg = RequestMessage(
            request_id=deserialize(blob=proto.request_id),
            status=proto.status,
            request_type=proto.request_type,
            date=proto.date,
            object_tags=proto.object_tags,
            object_type=proto.object_type,
            user_name=proto.user_name,
            user_email=proto.user_email,
            user_role=proto.user_role,
            requested_budget=proto.requested_budget,
            current_budget=proto.current_budget,
            request_description=proto.request_description,
            address=deserialize(blob=proto.target_address),
            object_id=deserialize(blob=proto.object_id),
            owner_address=deserialize(blob=proto.owner_address),
            requester_verify_key=VerifyKey(proto.requester_verify_key),
            timeout_secs=proto.timeout_secs,
        )
        request_msg.request_id = deserialize(blob=proto.request_id)
        return request_msg

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RequestMessage_PB



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/service/signaling_service.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.node.common import (
    metadata_pb2 as proto_dot_core_dot_node_dot_common_dot_metadata__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*proto/grid/service/signaling_service.proto\x12\x11syft.grid.service\x1a%proto/core/common/common_object.proto\x1a%proto/core/node/common/metadata.proto\x1a\x1bproto/core/io/address.proto"\x90\x01\n\x16RegisterNewPeerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"|\n\x1aPeerSuccessfullyRegistered\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07peer_id\x18\x03 \x01(\t"\xd8\x01\n\x16SignalingAnswerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07payload\x18\x03 \x01(\t\x12\x36\n\rhost_metadata\x18\x04 \x01(\x0b\x32\x1f.syft.core.node.common.Metadata\x12\x13\n\x0btarget_peer\x18\x05 \x01(\t\x12\x11\n\thost_peer\x18\x06 \x01(\t"\xd7\x01\n\x15SignalingOfferMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07payload\x18\x03 \x01(\t\x12\x36\n\rhost_metadata\x18\x04 \x01(\x0b\x32\x1f.syft.core.node.common.Metadata\x12\x13\n\x0btarget_peer\x18\x05 \x01(\t\x12\x11\n\thost_peer\x18\x06 \x01(\t"\xb9\x01\n\x17OfferPullRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x13\n\x0btarget_peer\x18\x04 \x01(\t\x12\x11\n\thost_peer\x18\x05 \x01(\t"\xba\x01\n\x18\x41nswerPullRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x13\n\x0btarget_peer\x18\x04 \x01(\t\x12\x11\n\thost_peer\x18\x05 \x01(\t"j\n\x19SignalingRequestsNotFound\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address"g\n\x16InvalidLoopBackRequest\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address"g\n\x16\x43loseConnectionMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.service.signaling_service_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _REGISTERNEWPEERMESSAGE._serialized_start = 173
    _REGISTERNEWPEERMESSAGE._serialized_end = 317
    _PEERSUCCESSFULLYREGISTERED._serialized_start = 319
    _PEERSUCCESSFULLYREGISTERED._serialized_end = 443
    _SIGNALINGANSWERMESSAGE._serialized_start = 446
    _SIGNALINGANSWERMESSAGE._serialized_end = 662
    _SIGNALINGOFFERMESSAGE._serialized_start = 665
    _SIGNALINGOFFERMESSAGE._serialized_end = 880
    _OFFERPULLREQUESTMESSAGE._serialized_start = 883
    _OFFERPULLREQUESTMESSAGE._serialized_end = 1068
    _ANSWERPULLREQUESTMESSAGE._serialized_start = 1071
    _ANSWERPULLREQUESTMESSAGE._serialized_end = 1257
    _SIGNALINGREQUESTSNOTFOUND._serialized_start = 1259
    _SIGNALINGREQUESTSNOTFOUND._serialized_end = 1365
    _INVALIDLOOPBACKREQUEST._serialized_start = 1367
    _INVALIDLOOPBACKREQUEST._serialized_end = 1470
    _CLOSECONNECTIONMESSAGE._serialized_start = 1472
    _CLOSECONNECTIONMESSAGE._serialized_end = 1575
# @@protoc_insertion_point(module_scope)



"""This module contains `Callable`, an AST node representing a method (can be static),
global function, or constructor which can be directly executed."""

# stdlib
from types import ModuleType
from typing import Any
from typing import Callable as CallableT
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# relative
from .. import ast
from .. import lib
from ..core.node.abstract.node import AbstractNodeClient
from ..core.node.common.action.function_or_constructor_action import (
    RunFunctionOrConstructorAction,
)
from ..logger import traceback_and_raise
from ..util import inherit_tags
from .util import module_type


class Callable(ast.attribute.Attribute):
    """Represent a method (can be static), global function, or constructor which can be directly executed."""

    def __init__(
        self,
        path_and_name: str,
        parent: ast.attribute.Attribute,
        object_ref: Optional[Any] = None,
        return_type_name: Optional[str] = None,
        client: Optional[AbstractNodeClient] = None,
        is_static: Optional[bool] = False,
    ):
        """Base constructor for Callable.

        Args:
            path_and_name: The path for the current node, e.g. `syft.lib.python.List`.
            parent: The parent node and it's attributes in the AST.
            object_ref: The actual python object for which the computation is being made.
            return_type_name: The return type name of the given action as a string with its full path.
            client: The client for which all computation is being executed.
            is_static: If True, the object has to be resolved on the AST, not on an existing pointer.
        """
        super().__init__(
            path_and_name=path_and_name,
            object_ref=object_ref,
            return_type_name=return_type_name,
            client=client,
            parent=parent,
        )

        self.is_static = is_static

    def __call__(
        self,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> Optional[Union["Callable", CallableT]]:
        """The `__call__` method on a `Callable` has two possible roles, e.g.

        1. If the client is set, execute the function for the client and return the appropriate pointer
        given the `return_type_name`.

        2. If the client is not set, then the `__call__` is used as a query on the ast.

        Args:
            *args: arguments of `callable`
            **kwargs: keyword arguments of `callable`

        Returns:
            If client is not set, returns `callable` node in AST at given path.
        """
        self.apply_node_changes()

        if self.client is not None:
            return_tensor_type_pointer_type = self.client.lib_ast.query(
                path=self.return_type_name
            ).pointer_type

            ptr = return_tensor_type_pointer_type(client=self.client)

            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=args, kwargs=kwargs)

            # then we convert anything which isn't a pointer into a pointer
            pointer_args, pointer_kwargs = ast.klass.pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            if self.path_and_name is not None:
                msg = RunFunctionOrConstructorAction(
                    path=self.path_and_name,
                    args=pointer_args,
                    kwargs=pointer_kwargs,
                    id_at_location=ptr.id_at_location,
                    address=self.client.address,
                    is_static=self.is_static,
                )

                self.client.send_immediate_msg_without_reply(msg=msg)

                inherit_tags(
                    attr_path_and_name=self.path_and_name,
                    result=ptr,
                    self_obj=None,
                    args=args,
                    kwargs=kwargs,
                )
                return ptr

        if "path" not in kwargs or "index" not in kwargs:
            traceback_and_raise(
                ValueError(
                    "AST with no client attached tries to execute remote function."
                )
            )
        path = kwargs["path"]
        index = kwargs["index"]

        if len(path) == index:
            return self.object_ref
        else:
            return self.attrs[path[index]](path=path, index=index + 1)

    def add_path(
        self,
        path: Union[str, List[str]],
        index: int,
        return_type_name: Optional[str] = None,
        framework_reference: Optional[ModuleType] = None,
        is_static: bool = False,
    ) -> None:
        """The add_path adds new nodes in AST based on type of current node and type of object to be added.

        Args:
            path: The node path added in AST, e.g. `syft.lib.python.List` or ["syft", "lib", "python", "List].
            index: The associated position in the path for the current node.
            return_type_name: The return type name of the given action as a string with its full path.
            framework_reference: The Python framework in which we can solve the same path to obtain the Python object.
            is_static: If the node represents a static method.
        """
        if index >= len(path) or path[index] in self.attrs:
            return

        attr_ref = getattr(self.object_ref, path[index])

        if isinstance(attr_ref, module_type):
            traceback_and_raise(
                ValueError("Module cannot be an attribute of Callable.")
            )

        self.attrs[path[index]] = ast.callable.Callable(
            path_and_name=".".join(path[: index + 1]),
            object_ref=attr_ref,
            return_type_name=return_type_name,
            client=self.client,
            parent=self,
        )



"""This module contains Attribute, an interface of a generic node in the AST."""

# stdlib
from types import ModuleType
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# relative
from .. import ast
from ..core.node.abstract.node import AbstractNodeClient
from ..logger import traceback_and_raise


class Attribute:
    """Attribute is the interface of a generic node in the AST that covers basic functionality."""

    __slots__ = [
        "path_and_name",
        "object_ref",
        "attrs",
        "return_type_name",
        "client",
        "_parent",
    ]

    lookup_cache: Dict[Any, Any] = {}

    def __init__(
        self,
        client: Optional[AbstractNodeClient],
        path_and_name: Optional[str] = None,
        object_ref: Any = None,
        return_type_name: Optional[str] = None,
        parent: Optional["Attribute"] = None,
    ):
        """Base constructor for all AST nodes.

        Args:
            client: The client for which all computation is being executed.
            path_and_name: The path for the current node, e.g. `syft.lib.python.List`.
            object_ref: The actual python object for which the computation is being made.
            return_type_name: The given action's return type name, with its full path, in string format.
            parent: The parent node in the AST.
        """
        self.client: Optional[AbstractNodeClient] = client
        self.path_and_name: Optional[str] = path_and_name
        self.object_ref: Any = object_ref
        self.return_type_name: Optional[str] = return_type_name

        # The `attrs` are the nodes that have the current node as a parent node
        # maps from the name on the path ot the actual attribute.
        self.attrs: Dict[str, "Attribute"] = {}
        self._parent: Optional["Attribute"] = parent

    def __call__(
        self,
        path: Union[List[str], str],
        index: int = 0,
        obj_type: Optional[type] = None,
    ) -> Any:
        """Execute the given node object reference with the given parameters.

        Args:
            path: The node path in AST to execute, e.g. `syft.lib.python.List` or ["syft", "lib", "python", "List].
            index: The associated position in the path for the current node.
            obj_type: The type of the object to be called, whose path is resolved from the `lookup_cache`.

        Returns:
            The results of running the computation on the object ref.
        """
        traceback_and_raise(NotImplementedError)

    def _extract_attr_type(
        self,
        container: Union[
            List["ast.klass.Class"],
            List["ast.module.Module"],
            List["ast.property.Property"],
        ],
        field: str,
    ) -> None:
        """Helper function to extract a class of nodes whose parent is the current node.

        Args:
            container: A list of objects in which we want to store the results.
            field: The typeof attribute from the current node's `attrs`.
        """
        for ref in self.attrs.values():
            sub_prop = getattr(ref, field, None)
            if sub_prop is None:
                continue

            container.extend(sub_prop)

    @property
    def classes(self) -> List["ast.klass.Class"]:
        """Extract all classes from the current node attributes.

        Returns:
            The list of classes in the current AST node attributes.
        """
        out = []

        if isinstance(self, ast.klass.Class):
            out.append(self)

        self._extract_attr_type(out, "classes")
        return out

    @property
    def properties(self) -> List["ast.property.Property"]:
        """Extract all properties from the current node attributes.

        Returns:
            The list of properties in the current AST node attributes.
        """
        out = []

        if isinstance(self, ast.property.Property):
            out.append(self)

        self._extract_attr_type(out, "properties")
        return out

    def query(
        self, path: Union[List[str], str], obj_type: Optional[type] = None
    ) -> "Attribute":
        """The query method is a tree traversal function based on the path to retrieve the node.

           It has a similar functionality to `__call__`,
           main difference being that `query` retrieves node without performing execution on node.

        Args:
            path: The node path in AST to be queried, e.g. `syft.lib.python.List` or ["syft", "lib", "python", "List"].
            obj_type: The type of the object that we want to call, whose path is resolved from the `lookup_cache`.

        Returns:
            The attribute in the AST at the given initial path.
        """
        # TODO: fix hacky work around
        if path == "syft.lib.python.list.List":
            path = "syft.lib.python.List"

        if obj_type is not None:
            # If the searched given type has already been seen, resolve it with the path from `lookup_cache`.
            if obj_type in self.lookup_cache:
                path = self.lookup_cache[obj_type]

        _path = path if isinstance(path, list) else path.split(".")

        if len(_path) == 0:
            return self

        # If the first element of the path is a child node, continue the query in the child node
        if _path[0] in self.attrs:
            return self.attrs[_path[0]].query(path=_path[1:])

        traceback_and_raise(
            ValueError(f"Path {'.'.join(_path)} not present in the AST.")
        )

    @property
    def name(self) -> str:
        """Retrieve the name of the current AST node from its `path_and_name`.

        Returns:
            The name of the current attribute.
        """
        path_and_name = self.path_and_name if self.path_and_name else ""
        return path_and_name.rsplit(".", maxsplit=1)[-1]

    def add_path(
        self,
        path: List[str],
        index: int,
        return_type_name: Optional[str] = None,
        framework_reference: Optional[ModuleType] = None,
        is_static: bool = False,
    ) -> None:
        """The add_path method adds new nodes in AST based on type of current node and type of object to be added.

        Args:
            path: The node path added in AST, e.g. `syft.lib.python.List` or ["syft", "lib", "python", "List].
            index: The associated position in the path for the current node.
            framework_reference: The Python framework in which we can resolve same path to obtain Python object.
            return_type_name: The return type name of the given action as a string with its full path.
            is_static: If the queried object is static, it has to be found on AST itself, not on an existing pointer.
        """
        traceback_and_raise(NotImplementedError)

    def fetch_live_object(self) -> Any:
        """Get the new object and its attributes from the client."""
        return getattr(self.parent.object_ref, self.name)

    def object_change(self) -> bool:
        """Check if client wants to change any nodes in the AST with a new object."""
        return id(self.fetch_live_object()) != id(self.object_ref)

    def reconstruct_node(self) -> None:
        """Changes node reference in the AST by adding the new object's reference as specified by client."""
        self.object_ref = self.fetch_live_object()

    def apply_node_changes(self) -> None:
        """Apply the changes in the nodes in the AST as specified by the client."""
        if self._parent and self.object_change():
            self.reconstruct_node()

    @property
    def parent(self) -> "Attribute":
        """Check if all the nodes have a parent node.

        Returns:
            Attribute: parent node

        Raises:
            AttributeError: If node has no parent attribute.
        """
        if self._parent:
            return self._parent

        raise AttributeError(f"Node {self} in the AST has not parent attribute set!")



# stdlib
from typing import Tuple

# third party
import numpy as np
from numpy.typing import NDArray

# relative
from ...common.serde.serializable import serializable


@serializable(recursive_serde=True)
class Initializer(object):
    __attr_allowlist__: Tuple[str] = ()  # type: ignore
    """Base class for parameter weight initializers.

    The :class:`Initializer` class represents a weight initializer used
    to initialize weight parameters in a neural network layer. It should be
    subclassed when implementing new types of weight initializers.
    """

    def __call__(self, size: Tuple[int, ...]) -> NDArray:
        """Makes :class:`Initializer` instances callable like a function, invoking
        their :meth:`call()` method.
        """
        return self.call(size)

    def call(self, size: Tuple[int, ...]) -> NDArray:
        """Sample should return a numpy.array of size shape and data type
        ``numpy.float32``.
        """
        raise NotImplementedError()

    def __str__(self) -> str:
        return self.__class__.__name__


@serializable(recursive_serde=True)
class Uniform(Initializer):
    __attr_allowlist__ = ("scale",)

    def __init__(self, scale: float = 0.05):
        self.scale = scale

    def call(self, size: Tuple[int, ...]) -> NDArray:
        return np.array(np.random.uniform(-self.scale, self.scale, size=size))


def decompose_size(size: Tuple[int, ...]) -> Tuple[int, int]:
    if len(size) == 2:
        fan_in = size[0]
        fan_out = size[1]

    elif len(size) == 4 or len(size) == 5:
        respective_field_size = np.prod(size[2:])
        fan_in = size[1] * respective_field_size
        fan_out = size[0] * respective_field_size

    else:
        fan_in = fan_out = int(np.sqrt(np.prod(size)))

    return fan_in, fan_out


@serializable(recursive_serde=True)
class XavierInitialization(Initializer):
    __attr_allowlist__: Tuple[str] = ()  # type: ignore

    def call(self, size: Tuple[int, ...]) -> NDArray:
        fan_in, fan_out = decompose_size(size)
        return Uniform(np.sqrt(6 / (fan_in + fan_out)))(size)



"""Strategies that clients could use to trigger the garbage collection process.

For the moment, there are implemented two GC strategies: GCSimple and GCBatched
Information about each strategy could be found in the implementation files.

By default, the strategy that is utilised is the "gcsimple" one.

The GCStrategy is on a per client basis.


1. How to use a specific strategy?

    client.gc = GarbageCollection(<strategy_name>, <args>, <kwargs>)

or

    client.gc.gc_strategy = <GCStrategyInstance>


2. How to implement a new strategy?
  - Create a new implementation file in the garbage_collection folder with a
  intuitive name
  - Implement the "reap" method (mandatory!) -- what should happen when a pointer
  object is deleted
  - Implement the "__del__" method (if necessary) -- this is primary used when
  a user would change the strategy on the fly
  - Add your method to the "__init__.py" file (this file)
  - Add tests for your new GCStrategy
  - And that's all!
"""
# relative
from .garbage_collection import GarbageCollection
from .gc_batched import GCBatched
from .gc_simple import GCSimple
from .gc_strategy import GCStrategy

__GC_DEFAULT_STRATEGY: str = "gcsimple"


def gc_get_default_strategy() -> str:
    """Retrieve the default garbage collection strategy.

    Return:
        A string representing the default strategy
    """
    return __GC_DEFAULT_STRATEGY


def gc_set_default_strategy(gc_strategy_name: str) -> None:
    """Set the default garbage collection strategy for new clients.

    Args:
        gc_strategy_name (str): the garbage collection strategy name

    Return:
        None
    """
    global __GC_DEFAULT_STRATEGY
    if gc_strategy_name not in GCStrategy.REGISTERED_GC_STRATEGIES:
        raise ValueError(f"{gc_strategy_name} not registered!")
    __GC_DEFAULT_STRATEGY = gc_strategy_name


__all__ = [
    "GarbageCollection",
    "GCSimple",
    "GCBatched",
    "get_default_strategy",
    "set_default_strategy",
]



# # stdlib
# import json
# from typing import Dict
# from typing import Optional
#
# # third party
# from google.protobuf.reflection import GeneratedProtocolMessageType
# from typing_extensions import final
#
# # syft absolute
# from syft import serialize
# from syft.core.common.message import ImmediateSyftMessageWithReply
# from syft.core.common.message import ImmediateSyftMessageWithoutReply
# from syft.core.common.serde.deserialize import _deserialize
# from syft.core.common.serde.serializable import serializable
# from syft.core.common.uid import UID
# from syft.core.io.address import Address
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     CreateWorkerMessage as CreateWorkerMessage_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     CreateWorkerResponse as CreateWorkerResponse_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     DeleteWorkerMessage as DeleteWorkerMessage_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     DeleteWorkerResponse as DeleteWorkerResponse_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     GetWorkerInstanceTypesMessage as GetWorkerInstanceTypesMessage_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     GetWorkerInstanceTypesResponse as GetWorkerInstanceTypesResponse_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     GetWorkerMessage as GetWorkerMessage_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     GetWorkerResponse as GetWorkerResponse_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     GetWorkersMessage as GetWorkersMessage_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     GetWorkersResponse as GetWorkersResponse_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     UpdateWorkerMessage as UpdateWorkerMessage_PB,
# )
# from syft.proto.grid.messages.infra_messages_pb2 import (
#     UpdateWorkerResponse as UpdateWorkerResponse_PB,
# )
#
#
# @serializable()
# @final
# class CreateWorkerMessage(ImmediateSyftMessageWithReply):
#     def __init__(
#         self,
#         address: Address,
#         content: Dict,
#         reply_to: Address,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
#         self.content = content
#
#     def _object2proto(self) -> CreateWorkerMessage_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: CreateWorkerMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return CreateWorkerMessage_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             content=json.dumps(self.content),
#             reply_to=serialize(self.reply_to),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: CreateWorkerMessage_PB,
#     ) -> "CreateWorkerMessage":
#         """Creates a CreateWorkerMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: CreateWorkerMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return CreateWorkerMessage(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             content=json.loads(proto.content),
#             reply_to=_deserialize(blob=proto.reply_to),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return CreateWorkerMessage_PB
#
#
# @serializable()
# @final
# class CreateWorkerResponse(ImmediateSyftMessageWithoutReply):
#     def __init__(
#         self,
#         address: Address,
#         status_code: int,
#         content: Dict,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id)
#         self.status_code = status_code
#         self.content = content
#
#     def _object2proto(self) -> CreateWorkerResponse_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: SignalingOfferMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return CreateWorkerResponse_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             status_code=self.status_code,
#             content=json.dumps(self.content),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: CreateWorkerResponse_PB,
#     ) -> "CreateWorkerResponse":
#         """Creates a SignalingOfferMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: SignalingOfferMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return CreateWorkerResponse(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             status_code=proto.status_code,
#             content=json.loads(proto.content),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return CreateWorkerResponse_PB
#
#
# @serializable()
# @final
# class GetWorkerInstanceTypesMessage(ImmediateSyftMessageWithReply):
#     def __init__(
#         self,
#         address: Address,
#         content: Dict,
#         reply_to: Address,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
#         self.content = content
#
#     def _object2proto(self) -> GetWorkerInstanceTypesMessage_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: GetWorkerInstanceTypesMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return GetWorkerInstanceTypesMessage_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             content=json.dumps(self.content),
#             reply_to=serialize(self.reply_to),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: GetWorkerInstanceTypesMessage_PB,
#     ) -> "GetWorkerInstanceTypesMessage":
#         """Creates a GetWorkerInstanceTypesMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: GetWorkerInstanceTypesMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return GetWorkerInstanceTypesMessage(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             content=json.loads(proto.content),
#             reply_to=_deserialize(blob=proto.reply_to),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return GetWorkerInstanceTypesMessage_PB
#
#
# @serializable()
# @final
# class GetWorkerInstanceTypesResponse(ImmediateSyftMessageWithoutReply):
#     def __init__(
#         self,
#         address: Address,
#         status_code: int,
#         content: Dict,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id)
#         self.status_code = status_code
#         self.content = content
#
#     def _object2proto(self) -> GetWorkerInstanceTypesResponse_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: SignalingOfferMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return GetWorkerInstanceTypesResponse_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             status_code=self.status_code,
#             content=json.dumps(self.content),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: GetWorkerInstanceTypesResponse_PB,
#     ) -> "GetWorkerInstanceTypesResponse":
#         """Creates a SignalingOfferMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: SignalingOfferMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return GetWorkerInstanceTypesResponse(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             status_code=proto.status_code,
#             content=json.loads(proto.content),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return GetWorkerInstanceTypesResponse_PB
#
#
# @serializable()
# @final
# class GetWorkerMessage(ImmediateSyftMessageWithReply):
#     def __init__(
#         self,
#         address: Address,
#         content: Dict,
#         reply_to: Address,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
#         self.content = content
#
#     def _object2proto(self) -> GetWorkerMessage_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: GetWorkerMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return GetWorkerMessage_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             content=json.dumps(self.content),
#             reply_to=serialize(self.reply_to),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: GetWorkerMessage_PB,
#     ) -> "GetWorkerMessage":
#         """Creates a GetWorkerMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: GetWorkerMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return GetWorkerMessage(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             content=json.loads(proto.content),
#             reply_to=_deserialize(blob=proto.reply_to),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return GetWorkerMessage_PB
#
#
# @serializable()
# @final
# class GetWorkerResponse(ImmediateSyftMessageWithoutReply):
#     def __init__(
#         self,
#         address: Address,
#         status_code: int,
#         content: Dict,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id)
#         self.status_code = status_code
#         self.content = content
#
#     def _object2proto(self) -> GetWorkerResponse_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: SignalingOfferMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return GetWorkerResponse_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             status_code=self.status_code,
#             content=json.dumps(self.content),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: GetWorkerResponse_PB,
#     ) -> "GetWorkerResponse":
#         """Creates a SignalingOfferMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: SignalingOfferMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return GetWorkerResponse(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             status_code=proto.status_code,
#             content=json.loads(proto.content),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return GetWorkerResponse_PB
#
#
# @serializable()
# @final
# class GetWorkersMessage(ImmediateSyftMessageWithReply):
#     def __init__(
#         self,
#         address: Address,
#         content: Dict,
#         reply_to: Address,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
#         self.content = content
#
#     def _object2proto(self) -> GetWorkersMessage_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: GetWorkersMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return GetWorkersMessage_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             content=json.dumps(self.content),
#             reply_to=serialize(self.reply_to),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: GetWorkersMessage_PB,
#     ) -> "GetWorkersMessage":
#         """Creates a GetWorkersMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: GetWorkersMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return GetWorkersMessage(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             content=json.loads(proto.content),
#             reply_to=_deserialize(blob=proto.reply_to),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return GetWorkersMessage_PB
#
#
# @serializable()
# @final
# class GetWorkersResponse(ImmediateSyftMessageWithoutReply):
#     def __init__(
#         self,
#         address: Address,
#         status_code: int,
#         content: Dict,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id)
#         self.status_code = status_code
#         self.content = content
#
#     def _object2proto(self) -> GetWorkersResponse_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: SignalingOfferMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return GetWorkersResponse_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             status_code=self.status_code,
#             content=json.dumps(self.content),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: GetWorkersResponse_PB,
#     ) -> "GetWorkersResponse":
#         """Creates a SignalingOfferMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: SignalingOfferMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return GetWorkersResponse(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             status_code=proto.status_code,
#             content=json.loads(proto.content),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return GetWorkersResponse_PB
#
#
# @serializable()
# @final
# class UpdateWorkerMessage(ImmediateSyftMessageWithReply):
#     def __init__(
#         self,
#         address: Address,
#         content: Dict,
#         reply_to: Address,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
#         self.content = content
#
#     def _object2proto(self) -> UpdateWorkerMessage_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: UpdateWorkerMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return UpdateWorkerMessage_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             content=json.dumps(self.content),
#             reply_to=serialize(self.reply_to),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: UpdateWorkerMessage_PB,
#     ) -> "UpdateWorkerMessage":
#         """Creates a UpdateWorkerMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: UpdateWorkerMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return UpdateWorkerMessage(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             content=json.loads(proto.content),
#             reply_to=_deserialize(blob=proto.reply_to),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return UpdateWorkerMessage_PB
#
#
# @serializable()
# @final
# class UpdateWorkerResponse(ImmediateSyftMessageWithoutReply):
#     def __init__(
#         self,
#         address: Address,
#         status_code: int,
#         content: Dict,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id)
#         self.status_code = status_code
#         self.content = content
#
#     def _object2proto(self) -> UpdateWorkerResponse_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: SignalingOfferMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return UpdateWorkerResponse_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             status_code=self.status_code,
#             content=json.dumps(self.content),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: UpdateWorkerResponse_PB,
#     ) -> "UpdateWorkerResponse":
#         """Creates a SignalingOfferMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: SignalingOfferMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return UpdateWorkerResponse(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             status_code=proto.status_code,
#             content=json.loads(proto.content),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return UpdateWorkerResponse_PB
#
#
# @serializable()
# @final
# class DeleteWorkerMessage(ImmediateSyftMessageWithReply):
#     def __init__(
#         self,
#         address: Address,
#         content: Dict,
#         reply_to: Address,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
#         self.content = content
#
#     def _object2proto(self) -> DeleteWorkerMessage_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: DeleteWorkerMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return DeleteWorkerMessage_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             content=json.dumps(self.content),
#             reply_to=serialize(self.reply_to),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: DeleteWorkerMessage_PB,
#     ) -> "DeleteWorkerMessage":
#         """Creates a DeleteWorkerMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: DeleteWorkerMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return DeleteWorkerMessage(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             content=json.loads(proto.content),
#             reply_to=_deserialize(blob=proto.reply_to),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return DeleteWorkerMessage_PB
#
#
# @serializable()
# @final
# class DeleteWorkerResponse(ImmediateSyftMessageWithoutReply):
#     def __init__(
#         self,
#         address: Address,
#         status_code: int,
#         content: Dict,
#         msg_id: Optional[UID] = None,
#     ):
#         super().__init__(address=address, msg_id=msg_id)
#         self.status_code = status_code
#         self.content = content
#
#     def _object2proto(self) -> DeleteWorkerResponse_PB:
#         """Returns a protobuf serialization of self.
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms the current object into the corresponding
#         Protobuf object so that it can be further serialized.
#         :return: returns a protobuf object
#         :rtype: SignalingOfferMessage_PB
#         .. note::
#             This method is purely an internal method. Please use serialize(object) or one of
#             the other public serialization methods if you wish to serialize an
#             object.
#         """
#         return DeleteWorkerResponse_PB(
#             msg_id=serialize(self.id),
#             address=serialize(self.address),
#             status_code=self.status_code,
#             content=json.dumps(self.content),
#         )
#
#     @staticmethod
#     def _proto2object(
#         proto: DeleteWorkerResponse_PB,
#     ) -> "DeleteWorkerResponse":
#         """Creates a SignalingOfferMessage from a protobuf
#         As a requirement of all objects which inherit from Serializable,
#         this method transforms a protobuf object into an instance of this class.
#         :return: returns an instance of SignalingOfferMessage
#         :rtype: SignalingOfferMessage
#         .. note::
#             This method is purely an internal method. Please use syft.deserialize()
#             if you wish to deserialize an object.
#         """
#
#         return DeleteWorkerResponse(
#             msg_id=_deserialize(blob=proto.msg_id),
#             address=_deserialize(blob=proto.address),
#             status_code=proto.status_code,
#             content=json.loads(proto.content),
#         )
#
#     @staticmethod
#     def get_protobuf_schema() -> GeneratedProtocolMessageType:
#         """Return the type of protobuf object which stores a class of this type
#         As a part of serialization and deserialization, we need the ability to
#         lookup the protobuf object type directly from the object type. This
#         static method allows us to do this.
#         Importantly, this method is also used to create the reverse lookup ability within
#         the metaclass of Serializable. In the metaclass, it calls this method and then
#         it takes whatever type is returned from this method and adds an attribute to it
#         with the type of this class attached to it. See the MetaSerializable class for
#         details.
#         :return: the type of protobuf object which corresponds to this class.
#         :rtype: GeneratedProtocolMessageType
#         """
#
#         return DeleteWorkerResponse_PB
